
(define eval (lambda (x e)
	(letrec

	; Default environment
	((initial-env
		'((#f.#f)
		(#t.#t)
		(bottom.bottom)
		(car.car)
		(cdr.cdr)
		(cond.cond)
		(cons.cons)
		(eq?.eq?)
		(lambda.lambda)
		(%lambda.%lambda)
		(letrec.letrec)
		(null?.null?)
		(pair?.pair?)
		(procedure?.procedure?)
		(quote.quote)
		(%void.%void)))

	; Evaluate each member of list X in environment E:
	; (M1 ... Mn) -> (eval[M1] ... eval[Mn])
	(evlist (lambda (x e)
		(cond ((null? x) '())
			(#t (cons (_eval (car x) e) (evlist (cdr x) e))))))

	; Find the value of a symbol
	(value (lambda (x e)
		(letrec ((b (assoc x e)))
			(cond ((null? b) '%void)
				(#t (cdr b))))))

	; Find the value of a symbol,
	; fail if the symbol is undefined
	(strict-value (lambda (x e)
		(letrec ((v (value x e)))
			(cond ((eq? v '%void) (bottom 'eval:symbol-not-bound: x))
				(#t v)))))

	; Create a LETREC* environment
	(makeenv (lambda (bindings env)
		(letrec

			; Fix recursive bindings. This is a mess.
			; Lexically scoped recursive procedures are
			; cyclic structures. The construction of such
			; structures requires mutation, which purely
			; applicative languages such as SketchyLISP do
			; not provide.
			; This implementation actually cheats: To fix
			; recursive bindings, all bindings of the symbols
			; bound in LETREC* are REMOVED from the lexical
			; contexts of (mutual) recursive functions.
			; This way, the function symbols are looked up
			; in the outer environment.
			; NOTE: this trick breaks tail call optimization.
			; E is the environment to be fixed.
			((fix-rec (lambda (e)
				(letrec

					; Names of symbols bound by LETREC*
					((names (map car e))

					; Remove bindings with the given NAMES from
					; the lexical context of the function F.
					; F-ENV is the context of F.
					(remove-names (lambda (f f-env)
						(letrec

							; The function without lexical context
							((func
								(list (car f) (cadr f) (caddr f)))

							(remove (lambda (env)
								(cond ((null? env) '())
									((member (caar env) names)
										(remove (cdr env)))
									(#t (cons (car env)
										(remove (cdr env))))))))

							(append func (list (remove f-env))))))

					; Clean up lexical contexts of lambda functions
					; bound by LETREC*. B is one binding of the context
					; created by LETREC*.
					(cleanup (lambda (b)
						(letrec ((sym (car b))
								(val (cdr b)))
							(cond ((not (pair? val)) val)
								((not (eq? (car val) '%lambda)) val)
								((null? (cdddr val)) val)
								(#t (cons sym
									(remove-names val (cadddr val))))))))

					; Fix recursive references in each binding of
					; LETREC*. B is the context created by LETREC*.
					(fix (lambda (b)
						(cond ((null? b) '())
							(#t (cons (cleanup (car b))
									(fix (cdr b))))))))

					(fix e))))

			; Evaluate and bind arguments.
			(make-bindings (lambda (bindings new-env)
				(cond ((null? bindings) new-env)
					(#t (make-bindings (cdr bindings)
						(cons (cons
								(caar bindings)
								; Evaluate arguments in the union of
								; the outer environment and the local
								; environment built so far.
								(_eval (cadar bindings)
									(append new-env env)))
							new-env)))))))

			(append (fix-rec (make-bindings bindings '())) env))))

	; Extract lexical environment of LAMBDA
	(lexical-env (lambda (x)
		(cond ((null? (cdddr x)) '())
			(#t (cadddr x)))))

	; Is symbol X bound in list A or environment E?
	(bound? (lambda (x a e)
		(cond ((eq? (assoc x e) '())
				(member x a))
			(#t #t))))

	; Create a new symbol.value pair
	(binding (lambda (sym env)
		(list (cons sym (value sym env)))))

	; Create closure from LAMBDA function
	(make-closure (lambda (args body env)
		(letrec ((make-env (lambda (expr new-env)
				(cond ((pair? expr)
						(make-env (cdr expr)
							(make-env (car expr) new-env)))
					((null? expr)
						new-env)
					((bound? expr args new-env)
						new-env)
					(#t 
						(append (binding expr env) new-env))))))
			(list '%lambda args body (make-env body '())))))

	; Select first clause of COND that has a true predicate.
	(do-cond (lambda (clauses env)
		(cond ((null? clauses)
				(bottom 'eval:cond-no-default))
			(#t (letrec ((tval (_eval (caar clauses) env)))
				(cond ((eq? tval #t)
						(cadar clauses))
					((eq? tval #f)
						(do-cond (cdr clauses) env))
					(#t (bottom 'eval:cond-no-tval tval))))))))


	; Is X a procedure?
	(procedure-p (lambda (x)
		(and (pair? x) (eq? (car x) '%lambda))))

	; Evaluate X in E
	(_eval (lambda (x e)
		(cond ((symbol? x)
				(strict-value x e))
			((procedure-p x)
				x)
			((boolean? x)
				x)
			((null? x)
				'())
			(#t (letrec ((fun (_eval (car x) e)))
				(cond ((eq? fun 'bottom)
						(list (append 'bottom: (cdr x))))
					((eq? fun 'car)
						(car (_eval (cadr x) e)))
					((eq? fun 'cdr)
						(cdr (_eval (cadr x) e)))
					((eq? fun 'cond)
						(_eval (do-cond (cdr x) e) e))
					((eq? fun 'cons)
						(cons (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq? fun 'eq?)
						(eq? (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq? fun 'letrec)
						(_eval (caddr x) (makeenv (cadr x) e)))
					((eq? fun 'lambda)
						(make-closure (cadr x) (caddr x) e))
					((eq? fun 'null?)
						(null? (_eval (cadr x) e)))
					((eq? fun 'pair?)
						(pair? (_eval (cadr x) e)))
					((eq? fun 'procedure?)
						(procedure-p (_eval (cadr x) e)))
					((eq? fun 'quote)
						(cadr x))
					((symbol? fun)
						(bottom (list 'eval:non-function fun)))
					((null? fun)
						(bottom (list 'eval:non-function (car x))))
					((eq? (car fun) '%lambda)
						(_eval (caddr fun)
							(append
								(pair (cadr fun) (evlist (cdr x) e))
								(lexical-env fun) e)))
					(#t (bottom (list 'eval:non-function fun))))))))))

	(_eval x (append e initial-env)))))

