/*
 * SketchyLISP -- An interpreter for a purely applicative dialect of Scheme
 * Copyright (C) 2005 Nils M Holm <nmh@t3x.org>
 * Derived from ArrowLISP, Copyright (C) 1998-2005 Nils M Holm.
 * See the file LICENSE for conditions of use.
 *
 * How to compile:
 *
 * cc -o sketchy -O sketchy.c
 * Flags:
 *    -DSMALL compiles a small memory version.
 *    -DSIGNAL redirects SIGINT to the error handler.
 *
 * Revision history:
 *
 * 2005-08-07
 *	- Added optional SIGINT handler.
 * 2005-08-07
 *	- Code cleanup.
 * 2005-08-05
 *	- Forked ANSI C version.
 */

#include <stdlib.h>
#ifdef __TURBOC__
 #include <io.h>
#else
 #include <unistd.h>
 #define setmode(fd, mode)
#endif
#include <stdio.h>
#ifdef SIGNAL
 #include <signal.h>
#endif
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#define MAJORVERSION	5
#define RELEASE		"2005-08-07"

#define NBASEREG	10	/* Number of Base Registers */
#define MAXWRAPCOL	255	/* Max. wrap column */

/*
 * Maximum number of nested load commands
 * (Recommended: 16-bit: 12, >=32-bit 16)
 */
#ifdef SMALL
#define	MAXLOADL	12
#else
#define	MAXLOADL	16
#endif

/* Maximum length for symbols and imploded lists */
#ifdef SMALL
#define	TEXTLEN		128
#else
#define	TEXTLEN		1024
#endif

/*
 * Number of nodes.
 * (Recommended: 16-bit: 12280, >=32-bit: 131072)
 */
#ifdef SMALL
#define	POOLSIZE	12280
#else
#define	POOLSIZE	131072
#endif

/*
 *	The NODE structure has a CAR and a CDR field.
 *	The use of each field is as follows:
 *	+---+---+--------------------------+
 *	| M | A |   14-bit Address Field   |
 *	+---+---+--------------------------+
 *	The sign bit M is used to tag used nodes during
 *	the mark step of the garbage collection.
 *	The A field is used to tag atoms: if A==1 in a
 *	CAR field, the node containing the field represents
 *	a character (used to name atoms). In this case,
 *	the CAR field contains the character itself in the
 *	LSB and A==1 in the MSB. CDR points to the next
 *	character or NIL. A is unused in CDR fields.
 *	The node pool is kept in two separate vectors
 *	called 'Car' and 'Cdr'. Consequently, Car[i] is
 *	used to access the CAR field of the i'th node in
 *	the pool and Cdr[i] is used to access its CDR
 *	field.
 *	The pointer reversing collector uses CDRs M flag
 *	to store the state of an already visited atom.
 */

/* Masks for the 'A' and 'M' flags as desribed above. */
#ifdef SMALL
#define	AFLAG	0x4000
#define	MFLAG	0x8000
#else
#define	AFLAG	0x20000000
#define	MFLAG	0x40000000
#endif

#define	EOT	-1	/* EOT indicator */
#define	DOT	-2	/* Internal: dot character */
#define	RPAREN	-3	/* Internal: right parenthesis */

/* Evaluator states */
#define	MATOM	'0' 	/* Processing Atom */
#define	MLIST	'1' 	/* Processing List */
#define	MBETA	'2' 	/* Beta-reducing */
#define	MBIND	'3' 	/* Processing bindings of LETREC */
#define	MLETR	'4' 	/* Finish LETREC */
#define	MCOND	'5' 	/* Processing predicates of COND */
#define	MCONJ	'6' 	/* Processing arguments of AND */
#define	MDISJ	'7' 	/* Processing arguments of OR */

struct counter {
	int	nred, nr1k, nr1m, nr1g;
};

int	NIL;			/* Not In List (or Pool) */
int	Car[POOLSIZE],		/* Vector holding CAR fields */
	Cdr[POOLSIZE];		/* Vector holding CDR fields */
char	*Infile;		/* Input file name */
char	*Buffer;		/* Ptr to current buffer */
int	Bp, Ep;			/* Input pointer, limit */
char	BufVec[MAXLOADL*TEXTLEN+MAXLOADL];	/* Load buffers */
int	Input;			/* Current input stream */
int	Rejected;		/* Unread character */
int	Mstack, Lstack;		/* Mode stack, List stack */
int	Bstack;			/* Binding stack, used by LETREC */
int	Estack;			/* Environment stack, for fixing closures */
int	Parent;			/* Parent pointer used in D/S/W GC */
int	Free;			/* Freelist */
int	Symbols;		/* Symbol table */
int	Packages;		/* Package list */
int	Digits[10];		/* Digit symbols */
int	SafeSymbols;		/* Safe copy of symbols */
int	Stack, Stack0;		/* Global stack, bottom of Stack */
int	Frame;			/* Current call frame */
int	Tmp, Tmp2;		/* Safe locations */
int	*Base[NBASEREG];	/* Base registers (roots for GC) */
int	Level;			/* Nesting level during input */
int	LoadLev;		/* Nesting level of LOAD meta cmds */
int	EvLev;			/* Number of nested EVALs */
int	ErrFlag;		/* Error flag */
int	Function;		/* Name of current lambda function */
int	Trace;			/* Name of traced function */
char	PrtBuf[MAXWRAPCOL+1];	/* Print buffer, */
int	Pp;			/* Pointer */
int	Quoted;			/* Quote flag of PRINT */
int	WrapCol;		/* Wrap column */
char	*Prefix;		/* Output line prefix */
int	MaxUse, LastUse;	/* Memory use gauge */
int	Line;			/* Input line number */
struct counter Reductions,	/* Reduction counter */
	Allocations;		/* Allocation counter */
int	StatFlag;		/* Statistics flag */
int	Ntrace;			/* Max functions to print in call trace */
int	ClPrLev;		/* Closure Print level */
int	LexEnv;			/* Environment for creating closures */
int	Bound;			/* Variables bound in a closure */
char	Image[TEXTLEN];		/* Name of initial load image */

/* Builtin symbol pointers (for fast lookup) */
int	S_and, S_apply, S_callcc, S_char, S_closure, S_cond,
	S_continuation, S_define, S_false, S_lambda, S_letrec,
	S_number, S_or, S_primitive, S_quote, S_special, S_string,
	S_t, S_true, S_void, S_last, S_0, S_1, S_2, S_3, S_4,
	S_5, S_6, S_7, S_8, S_9;

/* Command line flags */
int	O_batch;	/* Batch (non-interactive) mode */

/* Primitive function opcodes */
enum {	P_BOTTOM, P_CAR, P_CDR, P_CHAR_TO_INTEGER, P_CHAR_CI_EQP,
	P_CHAR_CI_LTP, P_CHAR_EQP, P_CHAR_LTP, P_CHARP, P_CONS,
	P_DLESS, P_DMINUS, P_DPLUS, P_EQP, P_INTEGER_TO_CHAR,
	P_INTEGER_TO_LIST, P_LIST_TO_INTEGER, P_LIST_TO_STRING, P_NULLP,
	P_NUMBERP, P_PACKAGE, P_PAIRP, P_PROCEDUREP, P_RECURSIVE_BIND,
	P_STRING_TO_LIST, P_STRING_TO_SYMBOL, P_STRINGP,
	P_SYMBOL_TO_STRING, P_SYMBOLP, P_VOID,
	N_PRIMITIVES };

/* Primitive function pointers. */
int	(*Primitives[N_PRIMITIVES])(int);

void	_print(int n);
void	_prnum(int n, int w, char *spaces);
int	_rdch(void);
int	addPackage(int sym);
int	addPrim(char *name, int opcode);
int	addSym(char *s, int v);
int	alloc(int pcar, int pcdr);
int	assoc(int x, int n);
int	atomic(int n);
int	badArgLst(int n);
int	bindArgs(int n, int name);
void	bindLetrec(int env);
void	bsave(int n);
int	bunsave(int k);
void	bye(void);
int	capture(void);
#ifdef SIGNAL
 void	catchIntr(int);
#endif
int	character(void);
int	ckCont(int n);
void	clearStats(void);
int	closure(int n);
void	collect(int n);
int	copy(int n);
int	copyBindings(void);
void	count(struct counter *c, int k);
int	countFree(void);
int	digitToValue(int n);
int	doApply(int n);
int	doBottom(int n);
int	doCallCC(int n);
int	doCar(int n);
int	doCdr(int n);
int	doCharCiEqP(int n);
int	doCharCiLtP(int n);
int	doCharEqP(int n);
int	doCharLtP(int n);
int	doCharP(int n);
int	doCharToInteger(int n);
int	doCons(int n);
int	doDLess(int n);
int	doDMinus(int n);
int	doDPlus(int n);
int	doDefine(int n);
int	doEqP(int n);
int	doIntegerToChar(int n);
int	doIntegerToList(int n);
int	doLambda(int n);
int	doListToInteger(int n);
int	doListToString(int n);
int	doNullP(int n);
int	doNumberP(int n);
int	doPackage(int n);
int	doPairP(int n);
int	doProcedureP(int n);
int	doQuote(int n);
int	doRecursiveBind(int n);
int	doStringP(int n);
int	doStringToList(int n);
int	doStringToSymbol(int n);
int	doSymbolP(int n);
int	doSymbolToString(int n);
int	doVoid(int n);
void	dump(char *p);
void	dumpState(char *s, int m);
void	dumpSymbols(int y);
int	error(char *m, int n);
int	eval(int n);
int	evalClause(int n);
int	evalLetrec(void);
int	explodeNum(char *s);
void	fatal(char *m);
int	findPackage(int sym);
int	findPsym(char *s, int y);
int	findSym(char *s);
int	finishLetrec(void);
void	fixClosures(void);
int	flatCopy(int n, int *lastp);
int	gc(void);
void	getCharArgs(int n, int *pc1, int *pc2, char *msg);
int	getName(char *p);
int	getPred(void);
void	getOpts(int argc, char **argv);
int	implodeStr(char *s);
void	init1(void);
void	init2(void);
void	mark(int n);
int	isAlist(int n);
int	isBound(int n);
int	lazyAtom(int n);
void	license(void);
int	load(char *p);
void	loadImage(char *p);
int	local(int n);
char	*locase(char *s);
void	lsave(int n);
int	lunsave(int k);
int	main(int argc, char **argv);
void	meta(void);
int	mkChar(int x);
int	mkLevEnv(int term, int locals);
void	msave(int v);
int	munsave(void);
void	newLine(void);
int	nextLetrec(int n);
void	nl(void);
int	numericStr(char *s);
void	pl(char *s);
void	pr(char *s);
void	prDepth(int n);
int	primitive(int *np);
void	print(int n);
int	printChar(int n);
int	printClosure(int n);
int	printCont(int n);
int	printNum(int n);
int	printPrim(int n);
int	printQuote(int n);
void	printStats(void);
int	printString(int n);
void	printTrace(int n);
void	printValue(struct counter *c, char *s);
void	prnum(int n, int w);
void	prznum(int n, int w);
int	quote(int n);
int	readList(void);
int	rdch(void); int c;
void	repl(void);
void	reset(struct counter *c);
void	resetState(void);
void	restoreBindings(int values);
int	resume(int n);
void	save(int n);
void	setFlag(char *name, int *intp, int val);
int	setupCond(int n);
int	setupLetrec(int n);
int	setupLogOp(int n);
int	skip(void);
int	stringLiteral(void);
int	symOrNum(int c);
int	tagged(int n);
void	tailCall(void);
void	traceCall(void);
void	unbindArgs(void);
int	unreadable(void);
int	unsave(int k);
void	updatePackages(int old, int new);
int	usage(void);
int	valueToDigit(int n);
int	wrongArgs(int n);
int	xread(void);
int	yesno(char *p, int *pf);

/* Emit a newline sequence */
void newLine(void) {
	write(1, "\n", 1);
}

/* Print the string S thru a buffered interface. */
void pr(char *s) {
	int	k, n, lim;

	if (!s) {
		if (Prefix && WrapCol)
			write(1, Prefix, strlen(Prefix));
		write(1, PrtBuf, Pp);
		Pp = 0;
		return;
	}
	k = strlen(s);
	lim = WrapCol? WrapCol: MAXWRAPCOL;
	while (Pp + k > lim) {
		if (Prefix && WrapCol)
			write(1, Prefix, strlen(Prefix));
		n = lim - Pp;
		memcpy(&PrtBuf[Pp], s, n);
		write(1, PrtBuf, lim);
		Pp = 0;
		k = k-n;
		s = &s[n];
		if (WrapCol) newLine();
	}
	if (k) {
		memcpy(&PrtBuf[Pp], s, k);
		Pp = Pp+k;
	}
}

/* Print a padded number with leading characters. */
void _prnum(int n, int w, char *spaces) {
	char	b[20];
	int	k;

	sprintf(b, "%d", n);
	k = strlen(b);
	if (k < w) pr(&spaces[k]);
	pr(b);
}

/* Print number with leading spaces. */
void prnum(int n, int w) {
	_prnum(n, w, "      ");
}

/* Print number with leading zeroes. */
void prznum(int n, int w) {
	_prnum(n, w, "000");
}

/* Flush the output buffer and emit a newline sequence. */
void nl(void) {
	pr(0);
	newLine();
}

/* Convert string to lower case. */
char *locase(char *s) {
	int	k, i;

	k = strlen(s)+1;
	for (i=0; i<k; i++)
		if ('A' <= s[i] && s[i] <= 'Z')
			s[i] = s[i] + ('a' - 'A');
	return s;
}

/* Print function names on call stack */
void traceCall(void) {
	int	s, n;

	s = Frame;
	n = Ntrace;
	while (s != NIL) {
		if (!n || Cdr[s] == NIL || Car[Cdr[s]] == NIL) break;
		if (n == Ntrace) pr("* Trace:");
		n = n-1;
		pr(" ");
		Quoted = 1;
		_print(Car[Cdr[s]]);
		s = Car[s];
	}
	if (n != Ntrace) nl();
}

/*
 * Print error message M and set ErrFlag.
 * If N != -1 print an additional line containing '* N'.
 */
int error(char *m, int n) {
	if (ErrFlag) return NIL;
	Prefix = "* ";
	pr("* ");
	if (Infile) {
		pr(locase(Infile));
		pr(": ");
	}
	if (Function != NIL) {
		Quoted = 1;
		_print(Function);
	}
	else {
		pr("Top-level");
	}
	pr("[");
	prnum(Line, 0);
	pr("]: ");
	pr(m);
	if (n != -1) {
		if (m[0]) pr(": ");
		Quoted = 1;
		_print(n);
	}
	nl();
	if (Frame != NIL) traceCall();
	if (O_batch) exit(1);
	ErrFlag = -1;
	Function = NIL;
	Prefix = 0;
	return NIL;
}

/* Print message M and halt the interpreter. */
void fatal(char *m) {
	error(m, -1);
	pr("* Fatal error, aborting");
	nl();
	exit(1);
}

/* Reset counter. */
void reset(struct counter *c) {
	c->nred = 0;
	c->nr1k = 0;
	c->nr1m = 0;
	c->nr1g = 0;
}

/* Increment counter. */
void count(struct counter *c, int k) {
	c->nred = c->nred+k;
	if (c->nred >= 1000) {
		c->nred = c->nred - 1000;
		c->nr1k = c->nr1k + 1;
		if (c->nr1k >= 1000) {
			c->nr1k = 0;
			c->nr1m = c->nr1m+1;
			if (c->nr1m >= 1000) {
				c->nr1m = 0;
				c->nr1g = c->nr1g+1;
				if (c->nr1g >= 1000)
					fatal("statistics: counter overflow");
			}
		}
	}
}

/* Print counter value. */
void printValue(struct counter *c, char *s) {
	if (c->nr1g) { prznum(c->nr1g, 0); pr(","); }
	if (c->nr1m) { prznum(c->nr1m, c->nr1g?3:0); pr(","); }
	if (c->nr1k) { prznum(c->nr1k, c->nr1m?3:0); pr(","); }
	prznum(c->nred, c->nr1k?3:0);
	pr(" ");
	pr(s);
	nl();
}

/*
 * Mark nodes which can be accessed through N.
 * This routine uses the Deutsch/Schorr/Waite algorithm
 * (aka pointer reversal algorithm) which marks the
 * nodes of a pool in constant space.
 * In addition to the Mark flag (MFLAG), it uses
 * another bit to save the state of the current
 * node (S). The (otherwise unused) MFLAG of CDR is
 * used to store the state.
 * Each visited node goes through these states:
 * M==0 S==0 unvisited, process CAR
 * M==1 S==1 CAR visited, process CDR
 * M==1 S==0 completely visited, return to parent
 */
void mark(int n) {
	int	p;

	Parent = NIL;	/* Initially, there is no parent node */
	while (1) {
		/* Reached a leaf? */
		if (n == NIL || Car[n] & MFLAG) {
			/* When the current node is a leaf and there is */
			/* no parent, the entire tree is marked. */
			if (Parent == NIL) break;
			if (Cdr[Parent] & MFLAG) {
				/* State 2: the CDR of the parent has */
				/* not yet been marked (M of CDR set). */
				/* Swap CAR and CDR pointers and */
				/* proceed with CDR. Set State==3. */
				p = Cdr[Parent] & ~MFLAG;
				Cdr[Parent] = Car[Parent] & ~MFLAG; /* S=0 */
				Car[Parent] = n | MFLAG;	/* M=1 */
				n = p;
			}
			else {
				/* State 3: CAR and CDR of parent done. */
				/* Return to the parent and restore */
				/* parent of paren */
				p = Parent;
				Parent = Cdr[p];
				Cdr[p] = n;
				n = p;
			}
		}
		else {
			/* State 1: The current node has not yet been */
			/* visited. */
			if (Car[n] & AFLAG) {
				/* If the node is an atom, go directly */
				/* to state 3: Save the parent in CDR, */
				/* make the current node the new parent */
				/* and move to its CDR. */
				p = Cdr[n];
				Cdr[n] = Parent;		/* S=0 */
				Parent = n;
				n = p;
				Car[Parent] = Car[Parent] | MFLAG; /* M=1 */
			}
			else {
				/* Go to state 2: like above, but save */
				/* the parent in CAR and proceed to CAR. */
				p = Car[n];
				Car[n] = Parent | MFLAG;	/* M=1 */
				Parent = n;
				n = p;
				Cdr[Parent] = Cdr[Parent] | MFLAG; /* S=1 */
			}
		}
	}
}

/*
 * Mark and Sweep Garbage Collection.
 * First, mark all nodes which can be accessed through
 * Base Registers (Base[]) and then reclaim untagged
 * nodes.
 */
int gc(void) {
	int	i, k;

	k = 0;
#ifdef DEBUG
	pr("GC called");
	nl();
#endif
	for (i=0; i<NBASEREG; i++) mark(Base[i][0]);
	Free = NIL;
	for (i=0; i<POOLSIZE; i++) {
		if (!(Car[i] & MFLAG)) {
			Cdr[i] = Free;
			Free = i;
			k = k+1;
		}
		else {
			Car[i] = Car[i] & ~MFLAG;
		}
	}
	if (MaxUse < POOLSIZE-k) MaxUse = POOLSIZE-k;
#ifdef DEBUG
	prnum(k, 0);
	pr(" nodes reclaimed");
	nl();
#endif
	return k;
}

/* Allocate a fresh node and initialize with PCAR,PCDR. */
int alloc(int pcar, int pcdr) {
	int	n;

	if (StatFlag) count(&Allocations, 1);
	if (Free == NIL) {
		gc();
		if (Free == NIL) fatal("out of nodes");
	}
	n = Free;
	Free = Cdr[Free];
	Car[n] = pcar;
	Cdr[n] = pcdr;
	return n;
}

/* Save node N on the Stack. */
void save(int n) {
	Tmp = n;	/* Otherwise, alloc() might recycle this node. */
	Stack = alloc(n, Stack);
	Tmp = NIL;
}

/*
 * Pop K nodes off the Stack and return
 * the most recently popped one.
 */
int unsave(int k) {
	int	n;

	while (k) {
		if (Stack == NIL) fatal("stack underflow");
		n = Car[Stack];
		Stack = Cdr[Stack];
		k = k-1;
	}
	return n;
}

/* Save value V on the M-Stack. */
void msave(int v) {
	/* Since the Mstack holds integer values rather than */
	/* nodes, the values are packaged in the character */
	/* fields of atoms. */
	Car[Mstack] = alloc(v|AFLAG, Car[Mstack]);
}

/* Pop a value off the M-Stack and return i */
int munsave(void) {
	int	v;

	if (Car[Mstack] == NIL) fatal("m-stack underflow");
	v = Car[Car[Mstack]] & ~AFLAG;		/* See msave() */
	Car[Mstack] = Cdr[Car[Mstack]];
	return v;
}

/* Save node N on the L-Stack. */
void lsave(int n) {
	Tmp = n;	/* Otherwise, alloc() might recycle this node. */
	Lstack = alloc(n, Lstack);
	Tmp = NIL;
}

/*
 * Pop K nodes off the L-Stack and return
 * the most recently popped one.
 */
int lunsave(int k) {
	int	n;

	while (k) {
		if (Lstack == NIL) fatal("l-stack underflow");
		n = Car[Lstack];
		Lstack = Cdr[Lstack];
		k = k-1;
	}
	return n;
}

/* Save node N on the B-Stack. */
void bsave(int n) {
	Tmp = n;	/* Otherwise, alloc() might recycle this node. */
	Bstack = alloc(n, Bstack);
	Tmp = NIL;
}

/*
 * Pop K nodes off the B-Stack and return
 * the most recently popped one.
 */
int bunsave(int k) {
	int	n;

	while (k) {
		if (Bstack == NIL) fatal("b-stack underflow");
		n = Car[Bstack];
		Bstack = Cdr[Bstack];
		k = k-1;
	}
	return n;
}

/*
 * Read a single character from the input stream
 * and return it. Rdch()==EOT indicates that the
 * input is exhausted.
 * Characters will be stripped to seven bits.
 */
int _rdch(void) {
	int	c;

	if (Rejected != EOT) {
		c = Rejected;
		Rejected = EOT;
		return c;
	}
	if (Bp >= Ep) {
		Ep = read(Input, Buffer, TEXTLEN);
		Bp = 0;
	}
	if (Ep < 1) return EOT;
	Bp = Bp+1;
	c = Buffer[Bp-1] & 127;
	if (c == '\n') Line = Line+1;
	return c;
}

/* Read a character and convert it to lower case. */
int rdch(void) { int c;
	c = _rdch();
	return tolower(c);
}

/*
 * Find a symbol named S in the symbol table Y.
 * Each symbol is represented by a (NAME.VALUE) pair
 * where NAME is a list of character nodes and value
 * may be any valid S-expression.
 * The symbol table is a list containing symbol
 * pairs ((N1.V1) ...).
 * When a symbol named S is found, return its
 * pair (S.V) and otherwise return NIL.
 */
int findPsym(char *s, int y) {
	int	n, i;

	while (y != NIL) {
		n = Car[Car[y]];
		i = 0;
		while (n != NIL && s[i]) {
			if (s[i] != (Car[n] & 255)) break;
			n = Cdr[n];
			i = i+1;
		}
		if (n == NIL && !s[i]) return Car[y];
		y = Cdr[y];
	}
	return NIL;
}

/*
 * Find the symbol S in the symbol table of any
 * package in the package list.
 */
int findSym(char *s) {
	int	p, y;

	/* First search the current package */
	y = findPsym(s, Symbols);
	if (y != NIL) return y;
	/* No match, search other packages. */
	p = Packages;
	while (p != NIL) {
		y = findPsym(s, Cdr[Car[p]]);
		if (y != NIL) return y;
		p = Cdr[p];
	}
	return NIL;
}

/* Implode a string. */
int implodeStr(char *s) {
	int	i, n, m, a;

	i = 0;
	if (s[i] == 0) return NIL;
	n = NIL;
	while (s[i]) {
		m = alloc(s[i] | AFLAG, NIL);
		if (n == NIL) {		/* Protect the first character */
			n = m;
			save(n);
		}
		else {			/* Just append the rest */
			Cdr[a] = m;
		}
		a = m;
		i = i+1;
	}
	unsave(1);
	return n;
}

/* Update symbol table pointer in package lis */
void updatePackages(int old, int new) {
	int	p;

	p = Packages;
	while (p != NIL) {
		if (Cdr[Car[p]] == old) {
			Cdr[Car[p]] = new;
			return;
		}
		p = Cdr[p];
	}
	if (Packages != NIL)
		fatal("symbol table not in package list? *BOOM*");
}

/*
 * Add the symbol S to the symbol table if it
 * does not already exis If it does exist,
 * return the existing symbol.
 * When adding a new symbol, initialize the
 * VALUE field with V. If V==0, make the symbol
 * a constant (bind it to itself).
 * Return the pair representing the symbol S.
 */
int addSym(char *s, int v) {
	int	n, m, osym;

	n = findSym(s);
	if (n != NIL) return n;
	n = implodeStr(s);
	save(n);
	m = alloc(n, v? v: n);
	save(m);
	osym = Symbols;
	Symbols = alloc(m, Symbols);
	unsave(2);
	updatePackages(osym, Symbols);
	return m;
}

/* Add primitive procedure. */
int addPrim(char *name, int opcode) {
	int	y;

	y = addSym(name, 0);
	Cdr[y] = alloc(S_primitive, NIL);
	Cdr[Cdr[y]] = alloc(opcode | AFLAG, NIL);
	Cdr[Cdr[Cdr[y]]] = y;
	return y;
}

/* Find a package. */
int findPackage(int sym) {
	int	p;

	p = Packages;
	while (p != NIL) {
		if (Car[Car[p]] == sym) return Car[p];
		p = Cdr[p];
	}
	return NIL;
}

/* Add a package. */
int addPackage(int sym) {
	int	y, p;

	y = findPackage(sym);
	if (y != NIL) return Cdr[y];
	p = alloc(sym, NIL);
	save(p);
	Packages = alloc(p, Packages);
	unsave(1);
	return Cdr[p];
}

/*
 * Read a list (S0 ... SN) and return (a pointer to) it.
 * This routine also recognizes pairs of the form (S0.S1).
 * For empty lists, it returns NIL.
 */
int readList(void) {
	int	n,	/* Node read */
		m,	/* Ptr to list */
		a,	/* Used to append nodes to m */
		c;	/* Member counter */
	char	*badpair;

	badpair = "bad pair";
	Level = Level+1;
	m = alloc(NIL, NIL);	/* Root node */
	save(m);
	a = NIL;
	c = 0;
	while (1) {
		if (ErrFlag) return NIL;
		n = xread();
		if (n == EOT) fatal("EOT in list");
		if (n == DOT) {
			if (c < 1) {
				error(badpair, -1);
				continue;
			}
			n = xread();
			Cdr[a] = n;
			if (xread() != RPAREN) {
				error(badpair, -1);
				continue;
			}
			unsave(1);
			Level = Level-1;
			return m;
		}
		if (n == RPAREN) break;
		if (a == NIL) 
			a = m;		/* First member: insert at root */
		else
			a = Cdr[a];	/* Following members: append */
		Car[a] = n;
		Cdr[a] = alloc(NIL, NIL); /* Alloc space for next member */
		c = c+1;
	}
	Level = Level-1;
	if (a != NIL) Cdr[a] = NIL;	/* Remove trailing empty node */
	unsave(1);
	return c? m: NIL;
}

/*
 * Read characters from the input device until a non-space
 * character [^\t ] is found. Return that character.
 */
int skip(void) {
	int	c;

	c = rdch();
	while (c == ' ' || c == '\t') c = rdch();
	return c;
}

/* Variables to dump to image file */
int *ImageVars[] = {
	&StatFlag, &ClPrLev, &WrapCol,
	&Symbols, &Packages, &Free,
	&S_and, &S_apply, &S_callcc, &S_char, &S_closure, &S_cond,
	&S_continuation, &S_define, &S_false, &S_letrec, &S_lambda,
	&S_number, &S_or, &S_primitive, &S_quote, &S_special, &S_string,
	&S_t, &S_true, &S_void, &S_last, &S_0, &S_1, &S_2, &S_3, &S_4,
	&S_5, &S_6, &S_7, &S_8, &S_9,
NULL };

/* Dump node pool image to file named in P. */
void dump(char *p) {
	int	fd, n, i;
	int	**v;
	char	magic[10];

	fd = open(p, O_CREAT | O_WRONLY);
	setmode(fd, O_BINARY);
	if (fd < 0) {
		error("cannot create file", -1);
		pr("* "); pr(p); nl();
		return;
	}
	strcpy(magic, "SKETCHY__");
	magic[7] = sizeof(int);
	magic[8] = MAJORVERSION;
	write(fd, magic, 9);
	n = POOLSIZE;
	write(fd, &n, sizeof(int));
	v = ImageVars;
	i = 0;
	while (v[i]) {
		write(fd, v[i], sizeof(int));
		i = i+1;
	}
	if (	write(fd, Car, POOLSIZE*sizeof(int)) != POOLSIZE*sizeof(int) ||
		write(fd, Cdr, POOLSIZE*sizeof(int)) != POOLSIZE*sizeof(int)
	) {
		close(fd);
		pr("dump failed"); nl();
		return;
	}
	close(fd);
	if (!O_batch) {
		pr("dumped"); nl();
	}
}

/*
 * Load S-expressions from the external file or device
 * named in the string P. Return nothing.
 */
int load(char *p) {
	int	o, b, e, r;
	char	*buf, *of;
	int	fd, oline;

	if (LoadLev+1 >= MAXLOADL) return error("load level too deep", -1);
	fd = open(p, O_RDONLY);
	if (fd < 0) {
		error("cannot open file", -1);
		pr("* "); pr(p); nl();
		return NIL;
	}
	LoadLev = LoadLev + 1;
	/* Save old I/O state */
	buf = Buffer;
	b = Bp;
	e = Ep;
	r = Rejected;
	/* Reset I/O buffer */
	Bp = 0;
	Ep = 0;
	Rejected = EOT;
	Buffer = &BufVec[(TEXTLEN+1)*LoadLev];
	/* Run the toplevel loop with redirected I/O */
	o = Input;
	Input = fd;
	oline = Line;
	Line = 1;
	of = Infile;
	Infile = p;
	repl();
	Infile = of;
	Line = oline;
	/* Restore previous I/O state */
	Rejected = r;
	Ep = e;
	Bp = b;
	Buffer = buf;
	LoadLev = LoadLev - 1;
	Input = o;
	close(fd);
	if (Level) error("unbalanced parentheses in loaded file", -1);
	if (!LoadLev && !O_batch) {
		pr("loaded"); nl();
	}
	return 0;
}

/*
 * Read all characters up to the next line terminator (or the EOT)
 * from the input device and store them in the vector P.
 * Convert each character read to lower case.
 * Return the terminating character.
 * This routine is used to accept parameters to meta commands.
 */
int getName(char *p) {
	int	c, i;

	c = skip();
	i = 0;
	while (c != ' ' && c != '\t' && c != '\n' && c != '\r' && c != EOT) {
		/* Lowcase characters to avoid */
		/* SUCH_NAMES under Unix. */
		c = tolower(c);
		p[i] = c;
		if (i >= TEXTLEN-2) {
			error("name too long", -1);
			return 0;
		}
		c = rdch();
		i = i+1;
	}
	p[i] = 0;
	return c;
}

/*
 * Extract a YES/NO, ON/OFF, 0/1 argument used to
 * enable and disable options.
 */
int yesno(char *p, int *pf) {
	int	c, f;

	c = getName(p);
	f = 2;
	if (p[0] == 'y' || p[0] == '1') f = -1;
	if (p[0] == 'n' || p[0] == '0') f = 0;
	if (p[0] == 'o') {
		if (p[1] == 'n') f = -1;
		if (p[1] == 'f') f = 0;
	}
	if (f == 2) error("Usage: command on/off", -1);
	pf[0] = f;
	return c;
}

/* Count the nodes in the freelis */
int countFree(void ) {
	int	n, k;

	k = 0;
	n = Free;
	while (n != NIL) {
		k = k+1;
		n = Cdr[n];
	}
	return k;
}

/* Print a line followed by a newline sequence. */
void pl(char *s) {
	pr(s); nl();
}


/* Print the conditions of use. */
void license(void) {
	int	w;

	w = WrapCol;
	WrapCol = 0;
pl(
"SketchyLISP -- An Interpreter for Purely Symbolic LISP");
pl(
"Copyright (C) 2005 Nils M Holm.  All rights reserved.");
pl("");
pl(
"Redistribution and use in source and binary forms, with or without");
pl(
"modification, are permitted provided that the following conditions");
pl(
"are met:");
pl(
"1. Redistributions of source code must retain the above copyright");
pl(
"   notice, this list of conditions and the following disclaimer.");
pl(
"2. Redistributions in binary form must reproduce the above copyright");
pl(
"   notice, this list of conditions and the following disclaimer in the");
pl(
"   documentation and/or other materials provided with the distribution.");
pl("");
pl(
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND");
pl(
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
pl(
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
pl(
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE");
pl(
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
pl(
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS");
pl(
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)");
pl(
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
pl(
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
pl(
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF");
pl(
"SUCH DAMAGE.");
pl("");
	WrapCol = w;
}

/* Normal termination */
void bye(void) {
	if (!O_batch) {
		pr("Bye"); nl();
	}
	exit(0);
}

/* Dump symbol table */
void dumpSymbols(int y) {
	int	p;

	Quoted = 1;
	p = Packages;
	pr("Packages:");
	while (p != NIL) {
		pr(" ");
		_print(Car[Car[p]]);
		if (Cdr[Car[p]] == Symbols) pr("[open]");
		p = Cdr[p];
	}
	nl();
	y = findPackage(y);
	_print(Car[y]);
	pr(": ");
	_print(Cdr[y]);
	nl();
}

/* Set and acknowledge state of flag */
void setFlag(char *name, int *intp, int val) {
	if (val != 2) {
		intp[0] = val;
		if (!LoadLev && !O_batch) {
			pr(name);
			pl(intp[0]? " on": " off");
		}
	}
}

/* Read and process meta (colon) commands. */
void meta(void) {
	int	c, k, k2, y;
	char	*q1, *q2;
	char	p[TEXTLEN+1];

	c = skip();
	if (c == 'c') {			/* Load conditionally */
		c = getName(p);
		q1 = strchr(p, '/');
		q2 = strchr(q1? q1: p, '.');
		if (q2 != NULL) *q2 = 0;
		y = findSym(locase(q1? &q1[1]: p));
		if (q2 != NULL) *q2 = '.';
		if (y == NIL || Cdr[y] == S_void) {
			load(locase(p));
		}
		else if (!LoadLev && !O_batch) {
			pr("found"); nl();
		}
	}
	else if (c == 'd') {		/* Dump node pool image */
		c = getName(p);
		dump(p);
	}
	else if (c == 'g') {		/* Trigger GC */
		k = countFree();
		k2 = gc();
		if (MaxUse != 0) LastUse = MaxUse;
		MaxUse = 0;
		prnum(LastUse, 6);
		pr(" nodes used max."); nl();
		prnum(k2-k, 6);
		pr(" nodes reclaimed"); nl();
		prnum(k2, 6);
		pr(" nodes in freelist"); nl();
	}
	else if (c == 'k') {		/* Switch closure printing on/off */
		c = getName(p);
		k = atoi(p);
		if (k > 2) k = 2;
		ClPrLev = k>2? 2: k;
		pr("closure print level = ");
		prnum(ClPrLev, 0); nl();
	}
	else if (c == 'l') {		/* Load program */
		c = getName(p);
		load(p);
	}
	else if (c == 'q') {		/* Quit */
		bye();
	}
	else if (c == 's') {		/* Switch statistics mode on/off */
		c = yesno(p, &k);
		setFlag("statistics", &StatFlag, k);
	}
	else if (c == 't') {		/* Specify/reset function to trace */
		c = getName(p);
		Trace = findSym(p);
		if (Trace != NIL) {
			if (!O_batch) {
				pr("tracing "); pr(p); nl();
			}
		}
		else if (p[0]) {
			error("no such symbol", -1);
			pr("* "); pr(p); nl();
		}
		else {
			if (!O_batch) {
				pr("trace off"); nl();
			}
		}
	}
	else if (c == 'v') {		/* Version info or License */
		c = getName(p);
		if (p[0]) {
			license();
		}
		else {
			pr("SketchyLISP (C) 2005 Nils M Holm"); nl();
			pr("Latest modification: ");
			pr(RELEASE); nl();
			prnum(POOLSIZE, 6);
			pr(" total nodes"); nl();
		}
	}
	else if (c == 'w') {		/* Specify wrap column */
		c = getName(p);
		k = atoi(p);
		if (!k) {
			WrapCol = 0;
			if (!O_batch) {
				pr("wrap=off"); nl();
			}
		}
		else if (k < 1 || k > MAXWRAPCOL) {
			error("bad wrap column", -1);
		}
		else {
			WrapCol = k;
			if (!O_batch) {
				pr("wrap="); prnum(WrapCol, 0); nl();
			}
		}
	}
	else if (c == 'y') {		/* Dump symbols */
		c = getName(p);
		dumpSymbols(findSym(p));
	}
	else if (c == '#') {		/* Reset line number */
		Line = 0;
	}
	else {
		pr("Usage: :c file  - load file conditionally"); nl();
		pr("       :d file  - dump image to file"); nl();
		pr("       :g       - collect garbage"); nl();
		pr("       :k 0/1/2 - set closure print level"); nl();
		pr("       :l file  - load file"); nl();
		pr("       :q       - quit"); nl();
		pr("       :s 0/1   - statistics on/off"); nl();
		pr("       :t name  - trace function (:t=off)"); nl();
		pr("       :v[l]    - show version/license"); nl();
		pr("       :w col   - set wrap column (0=none)"); nl();
		pr("       :y [sym] - dump symbol table"); nl();
	}
	while (c != '\n' && c != EOT) c = rdch();
}

/* Is N a 'real' (non-NIL) Atom? */
int atomic(int n) {
	return n != NIL && Car[n] != NIL && Car[Car[n]] & AFLAG;
}

/* Is N a tagged list (an internal type)? */
int tagged(int n) {
	if (n == NIL || Car[n] == S_true || Car[n] == S_false)
		return 0;
	n = Car[Car[Car[n]]];
	return n & AFLAG && (n & ~AFLAG) == '#';
}

/* Is N a lazy atom (atomic or tagged list)? */
int lazyAtom(int n) {
	return atomic(n) || tagged(n);
}

/* Quote an expression */
int quote(int n) {
	int	q;

	save(n);
	q = alloc(S_quote, NIL);
	save(q);
	Cdr[q] = alloc(n, NIL);
	unsave(2);
	return q;
}

/*
 * Check whether a string represents a number.
 * Numbers are defined as [+-]?[0-9]+.
 */
int numericStr(char *s) {
	int	i;

	i = 0;
	if (s[0] == '+' || s[0] == '-')
		i = 1;
	if (!s[i]) return 0;
	while (s[i]) {
		if (!isdigit(s[i])) return 0;
		i = i+1;
	}
	return 1;
}

/* Explode a string into a list of one-character symbols. */
int explodeNum(char *s) {
	int	i, l, x, y;
	char	name[3];

	i = 0;
	l = alloc(S_number, NIL);
	x = l;
	save(l);
	strcpy(name, "0d");
	while (s[i]) {
		name[0] = s[i];
		name[1] = isdigit(s[i])? 'd': 0;
		y = addSym(name, NIL);
		Cdr[x] = alloc(y, NIL);
		x = Cdr[x];
		i = i+1;
	}
	unsave(1);
	return l;
}

/* Report unreadable object */
int unreadable(void) {
	int	c;
	char	buf[2];

	error("unreadable object", -1);
	pr("* #<");
	buf[1] = 0;
	while (1) {
		c = rdch();
		if (c == '>' || c == '\n') break;
		buf[0] = c;
		pr(buf);
	}
	pr(">"); nl();
	return NIL;
}

/* Create a character literal. */
int mkChar(int x) {
	int	ch;

	Tmp2 = alloc(S_char, NIL);
	Cdr[Tmp2] = alloc(x | AFLAG, NIL);
	ch = Tmp2;
	Tmp2 = NIL;
	return ch;
}

/* Read a character literal. */
int character(void) {
	char	buf[10];
	int	i, c;

	for (i=0; i<9; i++) {
		c = _rdch();
		if (i > 0 && !isalpha(c)) break;
		buf[i] = c;
	}
	Rejected = c;
	buf[i] = 0;
	if (i == 0) c = ' ';
	else if (i == 1) c = buf[0];
	else if (!strcmp(buf, "space")) c = ' ';
	else return error("bad character name", -1);
	return mkChar(c);
}

/* Read a string literal. */
int stringLiteral(void) {
	char	s[TEXTLEN];
	int	c, i, n, q;

	i = 0;
	q = 0;
	c = _rdch();
	while (q || c != '"') {
		if (i >= TEXTLEN-2) {
			error("symbol too long", -1);
			i = i-1;
		}
		s[i] = c;
		q = c == '\\';
		if (!q) i = i+1;
		c = _rdch();
	}
	s[i] = 0;
	n = implodeStr(s);
	save(n);
	n = alloc(n, NIL);
	n = alloc(S_string, n);
	unsave(1);
	return n;
}

/*
 * Read a symbol or a numeric literal. When reading a
 * symbol, add it to the global symbol table.
 */
int symOrNum(int c) {
	char	s[TEXTLEN];
	int	i;

	i = 0;
	while (	c != ' ' && c != '\t' && c != '\n' &&
		c != '\r' && c != '(' && c != ')' &&
		c != ';' && c != '.' && c != '#' &&
		c != '\''
	) {
		if (i >= TEXTLEN-2) {
			error("symbol too long", -1);
			i = i-1;
		}
		s[i] = c;
		i = i+1;
		c = rdch();
	}
	s[i] = 0;
	Rejected = c;
	if (numericStr(s)) return explodeNum(s);
	return addSym(s, S_void);
}

/*
 * Read an expression from the current input stream
 * and return (a pointer to) it.
 * Each valid expression is either an Atom or a List.
 */
int xread(void) {
	int	c;

	c = rdch();
	while (1) {	/* Skip spaces and comments, run meta commands */
		while (c == ' ' || c == '\t' || c == '\n' || c == '\r')
			c = rdch();
		if (c == ':' && !Level) {
			meta();
			c = '\n';
			ErrFlag = 0;
			continue;
		}
		if (c != ';') break;
		while (c != '\n') c = rdch();
	}
	if (c == EOT) return EOT;
	if (c == '(') {
		return readList();
	}
	else if (c == '\'') {
		return quote(xread());
	}
	else if (c == '#') {
		c = rdch();
		if (c == 'f') return S_false;
		if (c == 't') return S_true;
		if (c == '\\') return character();
		if (c == '<') return unreadable();
		return error("bad # syntax", -1);
	}
	else if (c == '"') {
		return stringLiteral();
	}
	else if (c == ')') {
		if (!Level) return error("unexpected ')'", -1);
		return RPAREN;
	}
	else if (c == '.') {
		if (!Level) return error("unexpected '.'", -1);
		return DOT;
	}
	else {
		return symOrNum(c);
	}
}

/* Misc. error reporting and handling routines... */

int wrongArgs(int n) {
	return error("wrong argument count", n);
}

int badArgLst(int n) {
	return error("bad argument list", n);
}

/* Evaluate N=(CONS M M2) */
int doCons(int n) {
	int	m, m2;

	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL)
		return wrongArgs(n);
	m2 = Car[Cdr[m]];
	m = alloc(Car[m], m2);
	return m;
}

/* Evaluate N=(CAR M) */
int doCar(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	if (atomic(Car[m]) || Car[m] == NIL || tagged(Car[m]))
		return error("non-pair in 'car'", Car[m]);
	return Car[Car[m]];
}

/* Evaluate N=(CDR M) */
int doCdr(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	if (atomic(Car[m]) || Car[m] == NIL || tagged(Car[m])) {
		return error("non-pair in 'cdr'", Car[m]);
	}
	m = Cdr[Car[m]];
	return m;
}

/* Evaluate N=(CHAR?INTEGER M) */
int doCharToInteger(int n) {
	int	m, i, c;
	char	b[4];

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (atomic(m) || m == NIL || Car[m] != S_char)
		return error("non-character in 'char->integer'", m);
	c = Car[Cdr[m]] & ~AFLAG;
	i = 3;
	b[i] = 0;
	while (c || i == 4) {
		i = i-1;
		b[i] = c % 10 + '0';
		c = c / 10;
	}
	return explodeNum(&b[i]);
}

/* Get arguments of CHAR... primitives with type checking. */
void getCharArgs(int n, int *pc1, int *pc2, char *msg) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL) {
		wrongArgs(n);
		return;
	}
	if (atomic(Car[m]) || Car[Car[m]] != S_char) {
		error(msg, Car[m]);
		return;
	}
	pc1[0] = Car[Cdr[Car[m]]] & ~AFLAG;
	m = Cdr[m];
	if (atomic(Car[m]) || Car[Car[m]] != S_char)
		error(msg, Car[m]);
	pc2[0] = Car[Cdr[Car[m]]] & ~AFLAG;
}

/* Evaluate N=(CHAR-CI<? M1 M2) */
int doCharCiLtP(int n) {
	int	c1, c2;

	getCharArgs(n, &c1, &c2, "non-char in char-ci<?");
	return tolower(c1) < tolower(c2)? S_true: S_false;
}

/* Evaluate N=(CHAR-CI==? M1 M2) */
int doCharCiEqP(int n) {
	int	c1, c2;

	getCharArgs(n, &c1, &c2, "non-char in char-ci=?");
	return tolower(c1) == tolower(c2)? S_true: S_false;
}

/* Evaluate N=(CHAR<? M1 M2) */
int doCharLtP(int n) {
	int	c1, c2;

	getCharArgs(n, &c1, &c2, "non-char in char<?");
	return c1 < c2? S_true: S_false;
}

/* Evaluate N=(CHAR==? M1 M2) */
int doCharEqP(int n) {
	int	c1, c2;

	getCharArgs(n, &c1, &c2, "non-char in char=?");
	return c1 == c2? S_true: S_false;
}

/* Evaluate N=(CHAR? M) */
int doCharP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	return !atomic(m) && Car[m] == S_char ? S_true: S_false;
}

/* Return the value of a digit; */
int digitToValue(int n) {
	int	y;

	if (!atomic(n)) return -1;
	y = Car[n];
	if (	!isdigit(Car[y] & ~AFLAG) ||
		(Car[Cdr[y]] & ~AFLAG) != 'd' ||
		Cdr[Cdr[y]] != NIL
	)
		return -1;
	return (Car[y] & ~AFLAG) - '0';
}

/* Convert a value to a digit. */
int valueToDigit(int n) {
	return Digits[n];
}

/* Evaluate N=(D+ M1 M2 M3) */
int doDPlus(int n) {
	int	m, d1, d2, d3, r, c;

	m = Cdr[n];
	if (	m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] == NIL ||
		Cdr[Cdr[Cdr[m]]] != NIL
	)
		return wrongArgs(n);
	d1 = digitToValue(Car[m]);
	if (d1 < 0) return error("non-digit in d+", Car[m]);
	d2 = digitToValue(Car[Cdr[m]]);
	if (d2 < 0) return error("non-digit in d+", Car[Cdr[m]]);
	d3 = digitToValue(Car[Cdr[Cdr[m]]]);
	if (d3 < 0) return error("non-digit in d+", Car[Cdr[Cdr[m]]]);
	r = d1 + d2 + d3;
	c = r / 10;
	r = r % 10;
	return alloc(valueToDigit(r), valueToDigit(c));
}

/* Evaluate N=(D- M1 M2 M3) */
int doDMinus(int n) {
	int	m, d1, d2, d3, r, b;

	m = Cdr[n];
	if (	m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] == NIL ||
		Cdr[Cdr[Cdr[m]]] != NIL
	)
		return wrongArgs(n);
	d1 = digitToValue(Car[m]);
	if (d1 < 0) return error("non-digit in d-", Car[m]);
	d2 = digitToValue(Car[Cdr[m]]);
	if (d2 < 0) return error("non-digit in d-", Car[Cdr[m]]);
	d3 = digitToValue(Car[Cdr[Cdr[m]]]);
	if (d3 < 0) return error("non-digit in d-", Car[Cdr[Cdr[m]]]);
	r = d1 - d2 - d3;
	b = 0;
	while (r < 0) {
		r = r + 10;
		b = b + 1;
	}
	return alloc(valueToDigit(r), valueToDigit(b));
}

/* Evaluate N=(D< M1 M2) */
int doDLess(int n) {
	int	m, d1, d2;

	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL)
		return wrongArgs(n);
	d1 = digitToValue(Car[m]);
	if (d1 < 0) return error("non-digit in d<", Car[m]);
	d2 = digitToValue(Car[Cdr[m]]);
	if (d2 < 0) return error("non-digit in d<", Car[Cdr[m]]);
	return d1 < d2? S_true: S_false;
}

/* Evaluate N=(INTEGER?CHAR M) */
int doIntegerToChar(int n) {
	int		m, p;
	unsigned char	c;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (atomic(m) || m == NIL || Car[m] != S_number)
		return error("non-integer in 'integer->char'", m);
	p = Cdr[m];
	c = 0;
	while (p != NIL) {
		c = c * 10 + (Car[Car[Car[p]]] & ~AFLAG) - '0';
		p = Cdr[p];
	}
	if (c > 127)
		return error("value out of range in 'integer->char'", m);
	return mkChar(c);
}

/* Evaluate N=(INTEGER?LIST M) */
int doIntegerToList(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (atomic(m) || m == NIL || Car[m] != S_number)
		return error("non-integer in 'integer->list'", m);
	return Cdr[m];
}

/* Evaluate N=(LIST?INTEGER M) */
int doListToInteger(int n) {
	int	m, check, p, d;

	m = Cdr[n];
	if (m == NIL || (Cdr[m] != NIL && Cdr[Cdr[m]] != NIL))
		return wrongArgs(n);
	check = Cdr[Cdr[m]] != NIL;
	m = Car[m];
	if (atomic(m) || tagged(m))
		return error("non-list in 'list->integer'", m);
	if (m == NIL)
		return error("empty list in 'list->integer'", m);
	if (check) {
		p = m;
		d = Car[p];
		if (	atomic(d) &&
			((Car[Car[d]] & ~AFLAG) == '+' ||
			 (Car[Car[d]] & ~AFLAG) == '-') &&
			Cdr[Car[d]] == NIL
		)
			p = Cdr[p];
		while (p != NIL) {
			d = Car[p];
			if (	d != S_0 && d != S_1 && d != S_2 &&
				d != S_3 && d != S_4 && d != S_5 &&
				d != S_6 && d != S_7 && d != S_8 &&
				d != S_9
			)
				error(
				"non-digit in argument to 'list->integer'",
					d);

			p = Cdr[p];
		}
	}
	return alloc(S_number, m);
}

/* Evaluate N=(NULL M) */
int doNullP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	return Car[m] == NIL? S_true: S_false;
}

/* Evaluate N=(EQ? M M2) */
int doEqP(int n) {
	int	m;
	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL)
		return wrongArgs(n);
	return Car[m] == Car[Cdr[m]]? S_true: S_false;
}


/* Evaluate N=(LIST?STRING M)  */
int doListToString(int n) {
	int	m, a, ch, p;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (m == NIL) {
		p = alloc(S_string, NIL);
		save(p);
		Cdr[p] = alloc(NIL, NIL);
		unsave(1);
		return p;
	}
	if (lazyAtom(m)) return error("non-list in 'list->string'", m);
	p = alloc(NIL, NIL);
	save(p);
	a = p;
	while (m != NIL) {
		ch = Car[m];
		if (atomic(ch) || Car[ch] != S_char)
			return error("non-char in argument to 'list->string'",
				ch);
		Car[a] = Car[Cdr[ch]];
		m = Cdr[m];
		if (m != NIL) {
			Cdr[a] = alloc(NIL, NIL);
			a = Cdr[a];
		}
	}
	p = alloc(p, NIL);
	Car[Stack] = p;
	p = alloc(S_string, p);
	unsave(1);
	return p;
}

/* Evaluate N=(NUMBER? M) */
int doNumberP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	return atomic(m) || m == NIL? S_false:
		Car[m] == S_number? S_true: S_false;
}

/* Evaluate N=(PACKAGE [N1 [N2]]) */
int doPackage(int n) {
	int	m;

	m = Cdr[n];
	if (m != NIL && Cdr[m] != NIL)
		return wrongArgs(n);
	m = m == NIL? NIL: Car[m];
	Symbols = addPackage(m);
	return m;
}

/* Evaluate N=(PAIR? M) */
int doPairP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	return atomic(m) || tagged(m) || m == NIL? S_false:
		S_true;
}

/* Evaluate N=(PROCEDURE? M) */
int doProcedureP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (atomic(m) || m == NIL)
		return S_false;
	if (	Car[m] == S_closure ||
		Car[m] == S_continuation ||
		Car[m] == S_primitive
	)
		return S_true;
	return S_false;
}

/* Evaluate N=(STRING?LIST M)  */
int doStringToList(int n) {
	int	m, a, p, lst;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (atomic(m) || m == NIL || Car[m] != S_string)
		return error("non-string in 'string->list'", m);
	p = Car[Cdr[m]];
	if (p == NIL) return NIL;
	lst = alloc(NIL, NIL);
	save(lst);
	a = lst;
	while (p != NIL) {
		Car[a] = alloc(S_char, NIL);
		Cdr[Car[a]] = alloc(Car[p], NIL);
		p = Cdr[p];
		if (p != NIL) {
			Cdr[a] = alloc(NIL, NIL);
			a = Cdr[a];
		}
	}
	unsave(1);
	return lst;
}

/* Evaluate N=(STRING?SYMBOL M)  */
int doStringToSymbol(int n) {
	char	s[TEXTLEN+1];
	int	m, i, p;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (atomic(m) || m == NIL || Car[m] != S_string)
		return error("non-string in 'string->symbol'", m);
	if (Cdr[m] == NIL || Car[Cdr[m]] == NIL)
		return error("empty string in 'string->symbol'", m);
	p = Car[Cdr[m]];
	i = 0;
	while (p != NIL) {
		if (i >= TEXTLEN-2)
			fatal("symbol too long in 'string->symbol'");
		s[i] = Car[p] & ~AFLAG;
		i = i+1;
		p = Cdr[p];
	}
	s[i] = 0;
	return addSym(s, S_void);
}

/* Evaluate N=(STRING? M) */
int doStringP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	return !atomic(m) && Car[m] == S_string ? S_true: S_false;
}

/* Evaluate N=(SYMBOL?STRING M)  */
int doSymbolToString(int n) {
	int	m, p;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	if (!atomic(m))
		return error("non-symbol in 'symbol->string'", m);
	p = alloc(Car[m], NIL);
	save(p);
	p = alloc(S_string, p);
	unsave(1);
	return p;
}


/* Evaluate N=(SYMBOL? M)  */
int doSymbolP(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = Car[m];
	return atomic(m) && m != S_true && m != S_false ?
		S_true: S_false;
}

/* Extract clause from argument list of COND. */
/* Check the syntax of the clause. */
/* Return the predicate of the clause for */
/* evaluation in EVAL. */
int getPred(void) {
	int	e;

	e = Car[Car[Bstack]];
	if (	atomic(e) || e == NIL ||
		Cdr[e] == NIL || Cdr[Cdr[e]] != NIL
	)
		return error("bad clause in 'cond'", e);
	return Car[e];	/* predicate */
}

/*
 * Setup context for evaluation of (COND (P1 E1) ... (Pn En)).
 * The context consits of a list of clauses.
 * return the predicate of the first clause.
 */
int setupCond(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL) return wrongArgs(n);
	bsave(m);
	return getPred();
}

/*
 * Evaluate next clause of COND.
 * N is the value of the current predicate.
 * If N=*T*, return the expression of the predicate.
 * If N=*F*, return the predicate of the next clause.
 * When returning the expression of a predicate (N=*T*),
 * set the context on the Bstack to NIL to signal that
 * a true clause was found.
 * Clauses and predicates are returned to EVAL for
 * evaluation.
 */
int evalClause(int n) {
	int	e;

	e = Car[Bstack];
	if (n == S_false) {
		Car[Bstack] = Cdr[e];
		if (Car[Bstack] == NIL)
			return error("no default in 'cond'", -1);
		return getPred();
	}
	else {
		e = Car[Cdr[Car[e]]];
		Car[Bstack] = NIL;
		return e;
	}
}

/*
 * Setup context for evaluation of (AND ...) and (OR ...)
 * Return the first expression of the form.
 */
int setupLogOp(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL) return wrongArgs(n);
	bsave(m);
	return Car[m];
}

/*
 * Unbind the arguments of LAMBDA and LETREC.
 * See also BINDARGS().
 */
void unbindArgs(void) {
	int	v;

	Frame = unsave(1);
	Function = unsave(1);
	v = bunsave(1);		/* Caller's namelist */
	while (v != NIL) {
		Cdr[Car[v]] = unsave(1);
		v = Cdr[v];
	}
}

/*
 * Check if the symbol N is bound in the current lexical
 * environment.
 */
int isBound(int n) {
	int	b;

	b = Bound;
	while (b != NIL) {
		if (atomic(b)) return n == b;
		if (n == Car[b]) return 1;
		b = Cdr[b];
	}
	b = Car[LexEnv];
	while (b != NIL) {
		if (Car[Car[b]] == n) return 1;
		b =Cdr[b];
	}
	return 0;
}

/*
 * Recursively collect free variables and add their symbols
 * and values to the current lexical environmen
 */
void collect(int n) {
	if (n == NIL || n & AFLAG || Car[n] & AFLAG) return;
	if (Car[Car[n]] & AFLAG) {
		if (isBound(n)) return;
		Car[LexEnv] = alloc(NIL, Car[LexEnv]);
		Car[Car[LexEnv]] = alloc(n, Car[n] == Cdr[n]? n: Cdr[n]);
		return;
	}
	collect(Car[n]);
	collect(Cdr[n]);
}

/* Create lexical environmen */
int mkLevEnv(int term, int locals) {
	LexEnv = alloc(NIL, NIL);
	save(LexEnv);
	Bound = atomic(locals)? alloc(locals, NIL) : locals;
	collect(term);
	unsave(1);
	return Car[LexEnv];
}

/* Create a closure from a lambda expression. */
int closure(int n) {
	int	cl, env, args, term;

	if (ErrFlag) return NIL;
	cl = alloc(S_closure, NIL);
	save(cl);
	args = Car[Cdr[n]];
	Cdr[cl] = alloc(args, NIL);
	term = Car[Cdr[Cdr[n]]];
	Cdr[Cdr[cl]] = alloc(term, NIL);
	if (Cdr[Cdr[Cdr[n]]] == NIL) {
		env = mkLevEnv(term, args);
		save(env);
		if (env != NIL) {
			Cdr[Cdr[Cdr[cl]]] = alloc(env, NIL);
			Estack = alloc(env, Estack);
		}
		unsave(1);
	}
	else {
		Cdr[Cdr[Cdr[cl]]] = alloc(Car[Cdr[Cdr[Cdr[n]]]], NIL);
	}
	unsave(1);
	return cl;
}

/* Fix recursive bindings in closures. */
void fixClosures(void) {
	int	a, ee, e;

	if (ErrFlag || Estack == NIL) return;
	a = Car[Bstack];
	while (a != NIL) {
		ee = Estack;
		while (ee != NIL) {
			e = Car[ee];
			while (e != NIL) {
				if (Car[a] == Car[Car[e]]) {
					Cdr[Car[e]] = Cdr[Car[a]];
					break;
				}
				e = Cdr[e];
			}
			ee = Cdr[ee];
		}
		a = Cdr[a];
	}
	Estack = NIL;
}

/* Check if N is an alis */
int isAlist(int n) {
	if (atomic(n)) return 0;
	while (n != NIL) {
		if (lazyAtom(Car[n]) || !atomic(Car[Car[n]]))
			return 0;
		n = Cdr[n];
	}
	return -1;
}

/* Evaluate to (ASSOC X N) */
int assoc(int x, int n) {
	while (n != NIL) {
		if (Car[Car[n]] == x)
			return Car[n];
		n = Cdr[n];
	}
	return NIL;
}

/* Evaluate N=(RECURSIVE-BIND M) */
int doRecursiveBind(int n) {
	int	m, env, envp, b, v, le, lep, lb, p;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	env = Car[m];
	if (!isAlist(env))
		return error("non-alist in recursive-bind", env);
	envp = env;
	/* Traverse environment */
	while (envp != NIL) {
		b = Car[envp];
		v = Cdr[b];
		/* Is the current binding a non-closure? */
		if (	atomic(v) || tagged(v) ||
			Car[v] != S_lambda || Cdr[v] == NIL ||
			Cdr[Cdr[v]] == NIL || Cdr[Cdr[Cdr[v]]] == NIL
		) {
			envp = Cdr[envp];
			continue;
		}
		/* Extract local environment of closure */
		le = Car[Cdr[Cdr[Cdr[v]]]];
		/* Should not happen, but anyway... */
		if (!isAlist(le)) {
			envp = Cdr[envp];
			continue;
		}
		/* Fix recursive bindings */
		lep = le;
		while (lep != NIL) {
			lb = Car[lep];
			p = assoc(Car[lb], env);
			if (p != NIL) {
				Cdr[lb] = Cdr[p];
			}
			lep = Cdr[lep];
		}
		envp = Cdr[envp];
	}
	return env;
}

/*
 * Set up a context for processing N=(LETREC ((MA1 eval[MX2]) ...) MN)
 * Save
 * - the complete LETREC expression on the Bstack
 * - the environment on the Bstack
 * - a list of new bindings on the Bstack (initially empty)
 * - two copies of the saved name list on the Stack
 */
int setupLetrec(int n) {
	int	m;	/* Argument pointer */

	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL)
		return wrongArgs(n);
	m = Car[m];
	if (atomic(m)) return error("bad environment in 'letrec'", m);
	bsave(n);	/* save entire LETREC */
	bsave(m);	/* save environment */
	bsave(NIL);	/* list of bindings */
	bsave(NIL);	/* save empty name list */
	return m;
}

/*
 * Process one binding of a LETREC:
 * bind value to name, advance to next binding.
 * Return:
 * non-NIL - more bindings in environment
 * NIL     - last binding done
 */
int nextLetrec(int n) {
	int	m, p;

	m = Car[Cdr[Cdr[Bstack]]];	/* rest of environment */
	if (m == NIL) return NIL;
	p = Car[m];
	Tmp2 = n;
	Car[Cdr[Bstack]] = alloc(NIL, Car[Cdr[Bstack]]);
	Car[Car[Cdr[Bstack]]] = alloc(Car[p], n);
	Tmp2 = NIL;
	Car[Cdr[Cdr[Bstack]]] = Cdr[m];
	return Cdr[m];
}

/*
 * Evaluate value to bind inside a LETREC:
 * - check LETREC syntax
 * - save name to bind to
 * - save original binding of name
 * - return value for evaluation in EVAL
 */
int evalLetrec(void) {
	int	m, p, v;

	m = Car[Cdr[Cdr[Bstack]]];
	p = Car[m];
	/* Each binding must have the form (atom expr) */
	if (	atomic(p) || Cdr[p] == NIL || atomic(Cdr[p]) ||
		Cdr[Cdr[p]] != NIL || !atomic(Car[p])
	) {
		/* In case of an error, get rid of the */
		/* partial environmen */
		v = bunsave(1);
		bunsave(3);
		bsave(v);
		save(Function);
		save(Frame);
		unbindArgs();
		return error("bad binding in 'letrec'", p);
	}
	Car[Bstack] = alloc(Car[p], Car[Bstack]);	/* Save name */
	/* Evaluate the new value of the current symbol */
	return Car[Cdr[p]];
}

/* Establish the bindings of LETREC. */
void bindLetrec(int env) {
	int	e, b;

	e = env;
	while (e != NIL) {
		b = Car[e];
		save(Cdr[Car[b]]);	/* Save old value */
		Cdr[Car[b]] = Cdr[b];	/* Bind new value */
		e = Cdr[e];
	}
}

/*
 * Finish processing bindings of LETREC:
 * finish context and return term of LETREC.
 */
int finishLetrec(void) {
	int	m, v, b;

	Tmp2 = alloc(NIL, NIL);	/* Create safe storage */
	Cdr[Tmp2] = alloc(NIL, NIL);
	Cdr[Cdr[Tmp2]] = alloc(NIL, NIL);
	v = bunsave(1);
	b = bunsave(1);	/* get bindings */
	m = bunsave(2);	/* drop environment, get full LETREC */
	Car[Tmp2] = b;		/* protect b, m, v */
	Car[Cdr[Tmp2]] = m;
	Cdr[Cdr[Tmp2]] = v;
	bindLetrec(b);
	bsave(v);
	fixClosures();
	save(Function);			/* required by UNBINDARGS() */
	save(Frame);
	Tmp2 = NIL;
	return Car[Cdr[Cdr[m]]];	/* return term of LETREC */
}

/* Make symbol N local to the current package. */
int local(int n) {
	int	y, osym;

	y = Symbols;
	while (y != NIL) {
		if (n == Car[y]) return n;
		y = Cdr[y];
	}
	osym = Symbols;
	Symbols = alloc(NIL, Symbols);
	Car[Symbols] = alloc(Car[n], S_void);
	updatePackages(osym, Symbols);
	return Car[Symbols];
}

/*
 * Evaluate N=(DEFINE M eval[M2])
 * The name M already has been added to the
 * symbol table by READ().
 */
int doDefine(int n) {
	int	m, v, y;

	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL)
		return wrongArgs(n);
	y = Car[m];
	if (!atomic(y)) return error("bad symbol name", y);
	/* Protect the unevaluated expression */
	v = Car[Cdr[m]];
	save(v);
	/* If we are binding to a lambda expression, */
	/* add a null environmen */
	if (!atomic(v) && Car[v] == S_lambda) {
		if (	Cdr[v] != NIL && Cdr[Cdr[v]] != NIL &&
			Cdr[Cdr[Cdr[v]]] == NIL
		) {
			Cdr[Cdr[Cdr[v]]] = alloc(NIL, NIL);
		}
	}
	y = local(y);
	/* Evaluate and bind second argumen */
	Cdr[y] = eval(Car[Cdr[m]]);
	unsave(1);
	return Car[m];
}

/* Evaluate N=(QUOTE M) */
int doQuote(int n) {
	int	m;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	return (Car[m]);
}

/* Evaluate N=(APPLY M) */
int doApply(int n) {
	int	m, p;

	m = Cdr[n];
	if (m == NIL || Cdr[m] == NIL || Cdr[Cdr[m]] != NIL)
		return wrongArgs(n);
	p = Car[Cdr[m]];
	while (p != NIL) {
		if (atomic(p) || tagged(p)) return
			error("improper list in 'apply'", Car[Cdr[m]]);
		p = Cdr[p];
	}
	return alloc(Car[m], Car[Cdr[m]]);
}

/* Evaluate N=(BOTTOM ...) */
int doBottom(int n) {
	return error("", n);
}

/* Check LAMBDA syntax and create closure from lambda expression. */
int doLambda(int n) {
	int	m, a;

	m = Cdr[n];
	if (	m == NIL || Cdr[m] == NIL ||
		(Cdr[Cdr[m]] != NIL && Cdr[Cdr[Cdr[m]]] != NIL)
	)
		return wrongArgs(n);
	if (Cdr[Cdr[m]] != NIL && !isAlist(Car[Cdr[Cdr[m]]]))
		return error("bad environment in 'lambda'",
			Car[Cdr[Cdr[m]]]);
	if (!atomic(Car[m])) {
		if (tagged(Car[m])) return badArgLst(Car[m]);
		m = Car[m];
		a = m;
		while (m != NIL) {
			if (!atomic(Car[m])) return badArgLst(a);
			if (atomic(Cdr[m])) break;
			m = Cdr[m];
		}
	}
	return Car[n] == S_closure? n: closure(n);
}

/*
 * Create a flat copy of a list.
 * Store a reference to the last member
 * of the copy in lastp.
 */
int flatCopy(int n, int *lastp) {
	int	a, m, last;

	if (n == NIL) {
		lastp[0] = NIL;
		return NIL;
	}
	m = alloc(NIL, NIL);
	save(m);
	a = m;
	last = m;
	while (n != NIL) {
		Car[a] = Car[n];
		last = a;
		n = Cdr[n];
		if (n != NIL) {
			Cdr[a] = alloc(NIL, NIL);
			a = Cdr[a];
		}
	}
	unsave(1);
	lastp[0] = last;
	return m;
}

/*
 * Create a 2-level copy of the Lstack
 * and fix cyclic references. N = Lstack.
 */
int copy(int n) {
	int	m, a, last, unused;

	if (n == NIL) return NIL;
	m = alloc(NIL, NIL);
	a = m;
	save(m);
	while (n != NIL) {
		if (Cdr[n] == NIL || Cdr[Cdr[n]] == NIL)
			fatal("internal: unexpected NIL in COPY");
		Car[a] = flatCopy(Car[n], &last);
		Cdr[a] = alloc(NIL, NIL);
		Car[Cdr[a]] = flatCopy(Car[Cdr[n]], &unused);
		Cdr[Cdr[a]] = alloc(last, NIL);
		n = Cdr[Cdr[Cdr[n]]];
	}
	unsave(1);
	return m;
}

/*
 * Copy names and values of the symbol table
 * into an alist.
 */
int copyBindings(void) {
	int	y, p, ny, pk, q;

	pk = Packages;
	p = alloc(NIL, NIL);
	ny = p;
	q = NIL;
	save(p);
	while (pk != NIL) {
		y = Cdr[Car[pk]];
		while (y != NIL) {
			Car[p] = alloc(Car[y], Cdr[Car[y]]);
			y = Cdr[y];
			Cdr[p] = alloc(NIL, NIL);
			q = p;
			p = Cdr[p];
		}
		pk = Cdr[pk];
	}
	if (q != NIL) Cdr[q] = NIL;
	unsave(1);
	return Car[ny] == NIL? NIL: ny;
}

/* Restore values of the symbol table. */
void restoreBindings(int values) {
	int	b;

	while (values != NIL) {
		b = Car[values];
		Cdr[Car[b]] = Cdr[b];
		values = Cdr[values];
	}
}

/* Capture interpreter state, creating a continuation. */
int capture(void) {
	int	cont, x;

	cont = alloc(S_continuation, NIL);
	save(cont);
	Cdr[cont] = alloc(NIL, NIL);
	Car[Cdr[cont]] = alloc(Car[Mstack], NIL);
	Cdr[Cdr[cont]] = alloc(NIL, NIL);
	Car[Cdr[Cdr[cont]]] = copy(Lstack);
	Cdr[Cdr[Cdr[cont]]] = alloc(Cdr[Cdr[Stack]], NIL);
	Cdr[Cdr[Cdr[Cdr[cont]]]] = alloc(NIL, NIL);
	Car[Cdr[Cdr[Cdr[Cdr[cont]]]]] = flatCopy(Bstack, &x);
	Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]] = alloc(Estack, NIL);
	Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]]] = alloc(NIL, NIL);
	Car[Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]]]] = copyBindings();
	unsave(1);
	return cont;
}

/* Check the arguments of a continuation. */
int ckCont(int n) {
	if (	Cdr[n] == NIL || Cdr[Cdr[n]] == NIL ||
		Cdr[Cdr[Cdr[n]]] == NIL || Cdr[Cdr[Cdr[Cdr[n]]]] == NIL ||
		Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]] == NIL ||
		Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]]] == NIL ||
		Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]]]] != NIL
	)
		error("invalid continuation", -1);
	return n;
}

/* Re-activate a captured continuation N. */
int resume(int n) {
	int	m, state, x;

	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	ckCont(Car[n]);
	if (ErrFlag) return NIL;
	state = Cdr[Car[n]];
	Car[Mstack] = Car[Car[state]];
	Lstack = Car[Cdr[state]];
	Stack = Car[Cdr[Cdr[state]]];
	Bstack = flatCopy(Car[Cdr[Cdr[Cdr[state]]]], &x);
	Estack = Car[Cdr[Cdr[Cdr[Cdr[state]]]]];
	restoreBindings(Car[Cdr[Cdr[Cdr[Cdr[Cdr[state]]]]]]);
	return Car[m];
}

/* Evaluate N=(CALL/CC M) */
int doCallCC(int n) {
	int	m;

	if (EvLev > 1) {
		error("'call/cc' cannot be used here; this is a known bug",
		-1);
		return NIL;
	}
	m = Cdr[n];
	if (m == NIL || Cdr[m] != NIL) return wrongArgs(n);
	m = alloc(Car[m], NIL);
	save(m);
	Cdr[m] = alloc(capture(), NIL);
	unsave(1);
	return m;
}

/* Evaluate N=(VOID) */
int doVoid(int n) {
	if (Cdr[n] != NIL) return wrongArgs(n);
	return S_void;
}

/*
 * Check whether (CAR NP[0]) is a builtin symbol.
 * If it is one, run the appropriate routine, save
 * its result in NP[0], and return -1.
 * Return 0 if (CAR NP[0]) is not a builtin procedure.
 */
int primitive(int *np) {
	int	n, y;
	int	(*op)(int);

	n = np[0];
	y = Car[n];
	if (Car[y] != S_primitive || ErrFlag) return 0;
	op = Primitives[Car[Cdr[y]] & ~AFLAG];
	n = (*op)(n);
	np[0] = n;
	return -1;
}

/*
 * Bind the arguments of a LAMBDA function.
 * For a lambda application N=((LAMBDA (X1 ... Xn) S [ENV]) Y1 ... Yn)
 * this includes the following steps for j in {1,...,n}:
 *	1) save Xj in V
 *	2) save the value of Xj
 *	3) bind Xj to Yj
 * This routine results in  S' == S[X1/Y1] ... [Xn/Yn].
 * S?S' is performed by creating a new contex BINDARGS()
 * has no function resul It is used by EVAL() to perform
 * BETA reduction steps.
 */
int bindArgs(int n, int name) {
	int	fa,	/* Formal arg list */
		aa,	/* Actual arg list */
		e;	/* S, as above */
	int	env;	/* Optional lexical environment */
	int	p;
	int	at;	/* Atomic argument list flag */

	if (ErrFlag) return NIL;
	fa = Car[Cdr[Car[n]]];
	at = atomic(fa);
	aa = Cdr[n];
	p = Cdr[Cdr[Car[n]]];
	e = Car[p];
	env = Cdr[p] != NIL ? Car[Cdr[p]]: NIL;
	bsave(NIL);
	while ((fa != NIL && aa != NIL) || at) {
		if (!at) {
			/* Save name */
			Car[Bstack] = alloc(Car[fa], Car[Bstack]);
			save(Cdr[Car[fa]]);		/* Save value */
			Cdr[Car[fa]] = Car[aa];		/* Bind arg */
			fa = Cdr[fa];
			aa = Cdr[aa];
		}
		if (atomic(fa)) {	/* improper argument list */
			Car[Bstack] = alloc(fa, Car[Bstack]);	/* Save name */
			save(Cdr[fa]);	/* Save value */
			Cdr[fa] = aa;	/* Bind remaining arg list */
			fa = NIL;
			aa = NIL;
			break;
		}
	}
	while (env != NIL) {
		p = Car[env];
		Car[Bstack] = alloc(Car[p], Car[Bstack]);/* Save name */
		save(Cdr[Car[p]]);		/* Save value */
		Cdr[Car[p]] = Cdr[p];		/* Bind lex val */
		env = Cdr[env];
	}
	if (fa != NIL || aa != NIL) {
		wrongArgs(n);
		n = NIL;
	}
	else {
		n = e;
	}
	save(Function);
	Function = name;
	save(Frame);
	Frame = Stack;
	return n;
}

/*
 * Print application of traced function N in the form
 *	* (NAME A1 ... An)
 * PRINT() cannot be used because it would print NAME in
 * its expanded (LAMBDA...) form which is not desirable.
 */
void printTrace(int n) {
	Prefix = "+ ";
	pr("+ ");
	pr("(");
	Quoted = 1;
	_print(Trace);
	while (1) {
		n = Cdr[n];
		if (n == NIL) break;
		pr(" ");
		_print(Car[n]);
	}
	pr(")"); nl();
	Prefix = 0;
}

/* Print depth of stack N. */
void prDepth(int n) {
	int	k;

	k = 0;
	while (n != NIL) {
		n = Cdr[n];
		k = k+1;
	}
	prnum(k, 6); pr(" ");
}

/* Debugging: dump interpreter state */
void dumpState(char *s, int m) {
	pr(s); nl();
	pr("       ");
	pr("Mode        = "); prnum(m-'0',0); nl();
	prDepth(Lstack);
	pr("Lstack      = "); print(Lstack); nl();
	pr("       ");
	pr("Mstack      = "); print(Mstack); nl();
	prDepth(Stack);       pr("Car[Stack]  == "); print(Car[Stack]); nl();
	pr("       ");
	pr("Stack0      = "); prnum(Stack0,0); nl();
	prDepth(Bstack);
	pr("Bstack      = "); print(Bstack); nl();
}

/* Do tail call optimization. */
void tailCall(void) {
	int	m, y;

	m = Car[Mstack];
	/* Skip over callee's LETREC frames, if any */
	while (m != NIL && (Car[m] & ~AFLAG) == MLETR) {
		m = Cdr[m];
	}
	/* Parent not beta-reducing? Give up. */
	if (m == NIL || (Car[m] & ~AFLAG) != MBETA)
		return;
	/* Yes, this is a tail call: */
	/* - remove callee's LETREC frames. */
	/* - remove caller's call frame. */
	while (1) {
		Tmp2 = unsave(1); /* M */
		unbindArgs();
		unsave(1);
		y = munsave();
		save(Tmp2);
		Tmp2 = NIL;
		if (y == MBETA) break;
	}
}

/*
 * Evaluate the term N and return its normal form.
 * This is the heart of the interpreter:
 * An iterative EVAL function with tail-call optimization and call/cc.
 * For details on the reduction rules, refer to the manual.
 * Basically, EVAL() performs these steps:
 *	- If N is NIL, return NIL
 *	- If N is a symbol, return its binding
 *	- If N is a list (A1 ... An)
 *		- If A1 not in {COND,DEFINE,LETREC,LAMBDA,QUOTE}
 *			- Create a new list (A1' ... An')
 *				where Aj'==EVAL(Aj) for j in {1,...,n}
 *		- If A1' (or A1) is a primitive (builtin fn), apply it,
 *			resulting in N'
 *		- If A1 is a lambda function, BETA-reduce it,
 *			resulting in N'
 *	- Return N'
 */
int eval(int n) {
	int	m,	/* Result node */
		m2,	/* Root of result lists */
		a,	/* Used to append to result */
		cbn;	/* Use call-by-name/quotation in next iteration */
	int	mode,	/* Current state */
		cf;	/* Continue flag */
	int	nm;	/* name of function to apply */

	EvLev = EvLev + 1;
	save(Stack0);
	Stack0 = Stack;
	mode = MATOM;
	cf = 0;
	cbn = 0;
	while (!ErrFlag) {
		if (StatFlag) count(&Reductions, 1);
		if (n == NIL) {			/* () -> () */
			m = NIL;
		}
		else if (atomic(n)) {		/* Symbol -> Value */
			if (cbn) {
				m = n;
				cbn = 0;
			}
			else {
				m = Cdr[n] == Car[n]? n: Cdr[n];
				if (m == S_void)
					return error("symbol not bound", n);
			}
		}
		else if (Car[n] == S_closure ||
			Car[n] == S_primitive ||
			Car[n] == S_special ||
			Car[n] == S_continuation ||
			Car[n] == S_number ||
			Car[n] == S_char ||
			Car[n] == S_string ||
			cbn == 2
		) {
			m = n;
			cbn = 0;
		}
		else {				/* List (...) and Pair (X.Y) */
			/*
			 * This block is used to DESCEND into lists.
			 * The following nodes/variables will be saved:
			 *	1) the original list (on Stack)
			 *	2) the current state (on Mstack)
			 *	3) the root of the result list (on Lstack)
			 *	4) a ptr to the next free node
			 *		in the result list (on Lstack)
			 *	5) a ptr to the next member of
			 *		the original list (on Lstack)
			 */
			m = Car[n];
			if (atomic(Cdr[n])) {
				error("improper list in application", n);
				n = NIL;
			}
			save(n);	/* Save original list */
			msave(mode);
			/* Check call-by-name built-ins and flag */
			if (	m == S_cond || m == S_and || m == S_or ||
				m == S_define || m == S_letrec ||
				m == S_quote || m == S_closure ||
				m == S_continuation ||
				m == S_lambda || cbn
			) {
				if (cbn) cbn = 0;
				lsave(NIL);
				lsave(NIL);
				lsave(n);	/* Root of result list */
				n = NIL;
			}
			else {
				a = alloc(NIL, NIL);
				lsave(a);
				lsave(Cdr[n]);
				lsave(a);	/* Root of result list */
				n = Car[n];
			}
			mode = MLIST;
			continue;
		}
		/*
		 * The following loop is used to ASCEND back to the
		 * root of a list, thereby performing BETA reduction
		 * and creating result lists.
		 */
		while (1) if (mode == MBETA || mode == MLETR) {
			/* Finish BETA reduction */
			unbindArgs();
			unsave(1);	/* Original list */
			mode = munsave();
		}
		else if (mode == MLIST) {	/* Append to list, reduce */
			n = Car[Cdr[Lstack]];	/* Next member */
			a = Car[Cdr[Cdr[Lstack]]];	/* Place to append to */
			m2 = lunsave(1);	/* Root of result list */
			if (a != NIL)		/* Append new member */
				Car[a] = m;
			if (n == NIL) {		/* End of list */
				m = m2;
				lunsave(2);	/* Drop N,A */
 				/* Drop original list, remember first element */
				nm = Car[unsave(1)];
				save(m);	/* Save result */
				if (Trace == nm) printTrace(m);
				if (primitive(&m))	/* primitive fn? */
					;
				else if (Car[m] == S_quote) {
					/* QUOTE is special */
					m = doQuote(m);
				}
				else if (Car[m] == S_cond) {
					/* COND is special */
					n = setupCond(m);
					cf = 2;
					mode = MCOND;
				}
				else if (!atomic(Car[m]) &&
					Car[m] != NIL &&
					Car[Car[m]] == S_closure
				) {
					/* Application: */
					/*   reduce ((lambda...)...) */
					nm = atomic(nm)? nm: NIL;
					/* If the caller is also */
					/* BETA-reducing, */
					/* this is a TAIL application. */
					tailCall();
					bindArgs(m, nm);
					/* N=S of ((LAMBDA (...) S) ...) */
					n = Car[Cdr[Cdr[Car[m]]]];
					cf = 2;
					mode = MBETA;
				}
				else if (Car[m] == S_and || Car[m] == S_or) {
					/* AND and OR are special */
					if (Cdr[m] == NIL) {
						m = Car[m] == S_and?
							S_true: S_false;
					}
					else if (Cdr[Cdr[m]] == NIL) {
						n = Car[Cdr[m]];
						cf = 1;
					}
					else {
						n = setupLogOp(m);
						cf = 2;
						mode = Car[m] == S_and?
							MCONJ: MDISJ;
					}
				}
				else if (Car[m] == S_lambda) {
					/* LAMBDA is special */
					m = doLambda(m);
				}
				else if (Car[m] == S_letrec) {
					/* LETREC is special */
					if (setupLetrec(m) != NIL)
						n = evalLetrec();
					else
						n = NIL;
					cf = 2;
					mode = MBIND;
				}
				else if (Car[m] == S_callcc) {
					/* CALL/CC is special */
					n = doCallCC(m);
					cf = 1;
				}
				else if (!atomic(Car[m]) &&
					Car[m] != NIL &&
					Car[Car[m]] == S_continuation
				) {
					n = resume(m);
					cbn = 2;
					cf = 1;
				}
				else if (Car[m] == S_apply) {
					/* APPLY is special */
					n = doApply(m);
					cbn = 1;
					cf = 1;
				}
				else if (Car[m] == S_define) {
					/* DEFINE is special */
					m = doDefine(m);
				}
				else {
					error("application of non-function",
						nm);
					n = NIL;
				}
				if (cf != 2) {
					unsave(1);	/* Drop result */
					mode = munsave();
				}
				/* Continue this evaluation. */
				/* Leave the ASCENDING loop and descend */
				/* once more into N. */
				if (cf) break;
			}
			else {			/* Append to list */
				lsave(m2);
				/* Create space for next argument */
				Cdr[a] = alloc(NIL, NIL);
				Car[Cdr[Cdr[Lstack]]] = Cdr[a];
				Car[Cdr[Lstack]] = Cdr[n];
				n = Car[n];	/* Evaluate next member */
				break;
			}
		}
		else if (mode == MCOND) {
			n = evalClause(m);
			if (Car[Bstack] == NIL) {
				unsave(1);	/* Original list */
				bunsave(1);
				mode = munsave();
			}
			cf = 1;
			break;
		}
		else if (mode == MCONJ || mode == MDISJ) {
			Car[Bstack] = Cdr[Car[Bstack]];
			if (	(m == S_false && mode == MCONJ) || 
				(m != S_false && mode == MDISJ) ||
				Car[Bstack] == NIL
			) {
				unsave(1);	/* Original list */
				bunsave(1);
				mode = munsave();
				n = m;
				cbn = 1;
			}
			else if (Cdr[Car[Bstack]] == NIL) {
				n = Car[Car[Bstack]];
				unsave(1);	/* Original list */
				bunsave(1);
				mode = munsave();
			}
			else {
				n = Car[Car[Bstack]];
			}
			cf = 1;
			break;
		}
		else if (mode == MBIND) {
			if (nextLetrec(m) == NIL) {
				n = finishLetrec();
				mode = MLETR;
			}
			else {
				n = evalLetrec();
			}
			cf = 1;
			break;
		}
		else {	/* Atom */
			break;
		}
		if (cf) {	/* Continue evaluation if requested */
			cf = 0;
			continue;
		}
		if (Stack == Stack0) break;
	}
	while (Stack != Stack0) unsave(1);
	Stack0 = unsave(1);
	EvLev = EvLev - 1;
	return m;		/* Return the evaluated expr */
}

/* Print lists of digits in condensed format. */
int printNum(int n) {
	char	s[2];

	if (Car[n] != S_number) return 0;
	s[1] = 0;
	n = Cdr[n];
	while (1) {
		if (n == NIL) break;
		s[0] = Car[Car[Car[n]]] & 255;
		pr(s);
		n = Cdr[n];
	}
	return -1;
}

/* Print expressions of the form (QUOTE X) as 'X. */
int printQuote(int n) {
	if (	Car[n] == S_quote &&
		Cdr[n] != NIL &&
		Cdr[Cdr[n]] == NIL
	) {
		n = Car[Cdr[n]];
		if (n != S_true && n != S_false) pr("'");
		_print(n);
		return 1;
	}
	return 0;
}

/* Print a closure. */
int printClosure(int n) {
	if (	Car[n] == S_closure &&
		Cdr[n] != NIL && !atomic(Cdr[n]) &&
		Cdr[Cdr[n]] != NIL && !atomic(Cdr[Cdr[n]])
	) {
		Quoted = 1;
		pr("#<closure ");
		_print(Car[Cdr[n]]);
		if (ClPrLev > 0) {
			pr(" ");
			_print(Car[Cdr[Cdr[n]]]);
			if (ClPrLev > 1 && Cdr[Cdr[Cdr[n]]] != NIL) {
				pr(" ");
				_print(Car[Cdr[Cdr[Cdr[n]]]]);
			}
		}
		pr(">");
		return -1;
	}
	return 0;
}

/* Print a continuation. */
int printCont(int n) {
	if (Car[n] == S_continuation) {
		pr("#<continuation>");
		return -1;
	}
	else {
		return 0;
	}
}

/* Print a character. */
int printChar(int n) {
	char	b[2];
	int	c;

	if (Car[n] != S_char) return 0;
	pr("#\\");
	c = Car[Cdr[n]] & ~AFLAG;
	if (c == ' ') {
		pr("space");
	}
	else {
		b[1] = 0;
		b[0] = c;
		pr(b);
	}
	return -1;
}

/* Print a string. */
int printString(int n) {
	char	b[2];
	int	p;

	if (Car[n] != S_string) return 0;
	pr("\"");
	p = Car[Cdr[n]];
	b[1] = 0;
	while (p != NIL) {
		b[0] = Car[p] & ~AFLAG;
		if (b[0] == '"' || b[0] == '\\') pr("\\");
		pr(b);
		p = Cdr[p];
	}
	pr("\"");
	return -1;
}

/* Print a primitive function. */
int printPrim(int n) {
	if (Car[n] != S_primitive && Car[n] != S_special) return 0;
	_print(Cdr[Cdr[n]]);
	return -1;
}

/* Recursively print the term N. */
void _print(int n) {
	char	s[TEXTLEN+1];
	int	i;

	if (n == NIL) {
		if (!Quoted) {
			pr("'");
			Quoted = 1;
		}
		pr("()");
	}
	else if (Car[n] & AFLAG) {
		/* Characters are limited to the symbol table */
		pr("#<unprintable object>");
	}
	else if (atomic(n)) {

		if (!Quoted) {
			if (	n != S_true &&
				n != S_false &&
				n != S_number &&
				n != S_void
			)
				pr("'");
			Quoted = 1;
		}
		i = 0;		/* Symbol */
		n = Car[n];
		while (n != NIL) {
			s[i] = Car[n];
			if (i < TEXTLEN-2) i = i+1;
			n = Cdr[n];
		}
		s[i] = 0;
		pr(s);
	}
	else {	/* List */
		if (printNum(n)) return;
		if (printChar(n)) return;
		if (printString(n)) return;
		if (printCont(n)) return;
		if (printClosure(n)) return;
		if (printPrim(n)) return;
		if (!Quoted) {
			pr("'");
			Quoted = 1;
		}
		if (printQuote(n)) return;
		pr("(");
		while (n != NIL) {
			_print(Car[n]);
			n = Cdr[n];
			if (	n != NIL &&
				(atomic(n) || tagged(n))
			) {
				pr(".");
				_print(n);
				n = NIL;
			}
			if (n != NIL) pr(" ");
		}
		pr(")");
	}
}

/* Print quoted expression. */
void print(int n) {
	Quoted = 0;
	_print(n);
}

/* Reset interpreter state. */
void resetState(void) {
	Stack = NIL;
	Lstack = NIL;
	Bstack = NIL;
	Estack = NIL;
	Frame = NIL;
	Function = NIL;
	EvLev = 0;
}

#ifdef SIGNAL
void catchIntr(int sig) {
	error("interrupted", -1);
	ErrFlag = 1;
	signal(SIGINT, catchIntr);
}
#endif

/* Initialize interpreter variables. */
void init1(void) {
	/* Misc. variables */
	NIL = POOLSIZE;
	Level = 0;
	resetState();
	Symbols = NIL;
	Packages = NIL;
	SafeSymbols = NIL;
	Tmp = NIL;
	Tmp2 = NIL;
	LoadLev = 0;
	Trace = NIL;
	Pp = 0;
	WrapCol = 0;
	Prefix = 0;
	MaxUse = 0;
	LastUse = 0;
	StatFlag = 0;
	Ntrace = 10;
	ClPrLev = 0;
	Image[0] = 0;
	Line = 1;
	/* Initialize Freelist */
	Free = NIL;
	/* Clear input buffer */
	Infile = 0;
	Buffer = BufVec;
	Input = 0;
	Bp = 0;
	Ep = 0;
	Rejected = EOT;
#ifdef SIGNAL
	signal(SIGINT, catchIntr);
#endif
}

/*
 * Second stage of initialization:
 * protect registers from GC,
 * build the free list,
 * create builtin symbols.
 */
void init2(void) {
	int	core;

	/* Protect base registers */
	Base[0] = &Symbols;
	Base[1] = &Stack;
	Base[2] = &Mstack;
	Base[3] = &Lstack;
	Base[4] = &Bstack;
	Base[5] = &Estack;
	Base[6] = &Tmp;
	Base[7] = &Tmp2;
	Base[8] = &SafeSymbols;
	Base[9] = &Packages;
	/* Create builtin symbols */
	S_0 = addSym("0d", 0); /* First GC will be triggered HERE */
	S_1 = addSym("1d", 0);
	S_2 = addSym("2d", 0);
	S_3 = addSym("3d", 0);
	S_4 = addSym("4d", 0);
	S_5 = addSym("5d", 0);
	S_6 = addSym("6d", 0);
	S_7 = addSym("7d", 0);
	S_8 = addSym("8d", 0);
	S_9 = addSym("9d", 0);
	Digits[0] = S_0;
	Digits[1] = S_1;
	Digits[2] = S_2;
	Digits[3] = S_3;
	Digits[4] = S_4;
	Digits[5] = S_5;
	Digits[6] = S_6;
	Digits[7] = S_7;
	Digits[8] = S_8;
	Digits[9] = S_9;
	/* Tags (especially #<primitive> and #<special>) */
	/* must be defined before any primitives. */
	S_char = addSym("#<char>", 0);
	S_closure = addSym("#<closure>", 0);
	S_continuation = addSym("#<continuation>", 0);
	S_number = addSym("#<number>", 0);
	S_primitive = addSym("#<primitive>", 0);
	S_special = addSym("#<special>", 0);
	S_string = addSym("#<string>", 0);
	S_void = addSym("#<void>", 0);
	S_and = addSym("and", 0);
	S_apply = addSym("apply", 0);
	addPrim("bottom", P_BOTTOM);
	S_callcc = addSym("call/cc", 0);
	addPrim("car", P_CAR);
	addPrim("cdr", P_CDR);
	addPrim("char->integer", P_CHAR_TO_INTEGER);
	addPrim("char-ci=?", P_CHAR_CI_EQP);
	addPrim("char-ci<?", P_CHAR_CI_LTP);
	addPrim("char=?", P_CHAR_EQP);
	addPrim("char<?", P_CHAR_LTP);
	addPrim("char?", P_CHARP);
	S_cond = addSym("cond", 0);
	addPrim("cons", P_CONS);
	addPrim("d+", P_DPLUS);
	addPrim("d-", P_DMINUS);
	addPrim("d<", P_DLESS);
	S_define = addSym("define", 0);
	addPrim("eq?", P_EQP);
	S_false = addSym("#f", 0);
	addPrim("integer->char", P_INTEGER_TO_CHAR);
	addPrim("integer->list", P_INTEGER_TO_LIST);
	S_lambda = addSym("lambda", 0);
	S_letrec = addSym("letrec", 0);
	addPrim("list->integer", P_LIST_TO_INTEGER);
	addPrim("list->string", P_LIST_TO_STRING);
	addPrim("null?", P_NULLP);
	addPrim("number?", P_NUMBERP);
	S_or = addSym("or", 0);
	addPrim("package", P_PACKAGE);
	addPrim("pair?", P_PAIRP);
	addPrim("procedure?", P_PROCEDUREP);
	S_quote = addSym("quote", 0);
	addPrim("procedure?", P_PROCEDUREP);
	addPrim("recursive-bind", P_RECURSIVE_BIND);
	addPrim("string->symbol", P_STRING_TO_SYMBOL);
	addPrim("string->list", P_STRING_TO_LIST);
	addPrim("string?", P_STRINGP);
	addPrim("symbol->string", P_SYMBOL_TO_STRING);
	addPrim("symbol?", P_SYMBOLP);
	addPrim("void", P_VOID);
	S_true = addSym("#t", 0);
	S_last = addSym("**", 0);
	Mstack = alloc(NIL, NIL);
	Primitives[P_BOTTOM] = &doBottom;
	Primitives[P_CAR] = &doCar;
	Primitives[P_CDR] = &doCdr;
	Primitives[P_CHAR_TO_INTEGER] = &doCharToInteger;
	Primitives[P_CHAR_CI_EQP] = &doCharCiEqP;
	Primitives[P_CHAR_CI_LTP] = &doCharCiLtP;
	Primitives[P_CHAR_EQP] = &doCharEqP;
	Primitives[P_CHAR_LTP] = &doCharLtP;
	Primitives[P_CHARP] = &doCharP;
	Primitives[P_CONS] = &doCons;
	Primitives[P_DPLUS] = &doDPlus;
	Primitives[P_DMINUS] = &doDMinus;
	Primitives[P_DLESS] = &doDLess;
	Primitives[P_EQP] = &doEqP;
	Primitives[P_INTEGER_TO_CHAR] = &doIntegerToChar;
	Primitives[P_INTEGER_TO_LIST] = &doIntegerToList;
	Primitives[P_LIST_TO_INTEGER] = &doListToInteger;
	Primitives[P_LIST_TO_STRING] = &doListToString;
	Primitives[P_NULLP] = &doNullP;
	Primitives[P_NUMBERP] = &doNumberP;
	Primitives[P_PACKAGE] = &doPackage;
	Primitives[P_PAIRP] = &doPairP;
	Primitives[P_PROCEDUREP] = &doProcedureP;
	Primitives[P_RECURSIVE_BIND] = &doRecursiveBind;
	Primitives[P_STRING_TO_LIST] = &doStringToList;
	Primitives[P_STRING_TO_SYMBOL] = &doStringToSymbol;
	Primitives[P_STRINGP] = &doStringP;
	Primitives[P_SYMBOL_TO_STRING] = &doSymbolToString;
	Primitives[P_SYMBOLP] = &doSymbolP;
	Primitives[P_VOID] = &doVoid;
	core = addSym("core", 0);
	Packages = alloc(core, Symbols);
	Packages = alloc(Packages, NIL);
	Symbols = addPackage(NIL);
}

/* Reset the reduction counter */
void clearStats(void) {
	reset(&Reductions);
	reset(&Allocations);
}

/* Print the number of reductions done in last EVAL */
void printStats(void) {
	printValue(&Reductions, "calls to eval");
	printValue(&Allocations, "nodes allocated");
}

/* This is the top-level Read-Eval-Print loop. */
void repl(void) {
	int	n;

	while (1) {
		ErrFlag = 0;
		n = xread();
		if (n == EOT) break;
		save(n);
		SafeSymbols = copyBindings();
		if (StatFlag) clearStats();
		n = eval(Car[Stack]);
		unsave(1);
		if (!ErrFlag) {
			Cdr[S_last] = n;
			if (!LoadLev) {
				if (!O_batch) pr("=> ");
				print(n); nl();
			}
			if (Stack != NIL)
				fatal("internal: unbalanced stack");
			if (StatFlag && !LoadLev) printStats();
		}
		else {
			restoreBindings(SafeSymbols);
		}
		resetState();
		while (Car[Mstack] != NIL) munsave();
	}
}

/* Load initial node pool image */
void loadImage(char *p) {
	int	fd, n, i;
	char	buf[10];
	int	**v;

	fd = open(p, O_RDONLY);
	setmode(fd, O_BINARY);
	if (fd < 0) {
		error("cannot open image", -1);
		pr("* "); pr(p); nl();
		exit(1);
	}
	read(fd, buf, 9);
	read(fd, &n, sizeof(int));
	if (	memcmp(buf, "SKETCHY", 7) ||
		buf[7] != sizeof(int) ||
		buf[8] != MAJORVERSION ||
		n != POOLSIZE
	) {
		error("bad image", -1);
		pr("* "); pr(p); nl();
		exit(1);
	}
	v = ImageVars;
	i = 0;
	while (v[i]) {
		read(fd, v[i], sizeof(int));
		i = i+1;
	}
	if (	read(fd, Car, POOLSIZE*sizeof(int)) != POOLSIZE*sizeof(int) ||
		read(fd, Cdr, POOLSIZE*sizeof(int)) != POOLSIZE*sizeof(int)
	) {
		error("bad image", -1);
		pr("* "); pr(p); nl();
		exit(1);
	}
	close(fd);
}

/* Print usage and fail. */
int usage(void) {
	error("Usage: sketchy [-b] [image]", -1);
	exit(1);
}

/* Evaluate the command line options */
void getOpts(int argc, char **argv) {
	char	*a;
	int	i, j, k;

	O_batch = 0;
	i = 1;
	while (i < argc) {
		k = strlen(argv[i]);
		a = argv[i];
		if (a[0] != '-') break;
		for (j=1; j<k; j++)
			if (a[j] == 'b' || a[j] == 'B') O_batch = 1;
			else usage();
		i = i+1;
	}
	if (i < argc) strcpy(Image, a);
}

int main(int argc, char **argv) {
	init1();
	getOpts(argc, argv);
	init2();
	if (!O_batch) {
		pr("SketchyLISP ");
		pr(RELEASE);
		pr(" (C) 2005 Nils M Holm"); nl();
	}
	if (Image[0]) {
		loadImage(Image);
	}
	else if (!O_batch) {
		pr("Warning: no image loaded"); nl();
	}
	repl();
	bye();
	return 0;
}

